# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
import os
import os.path as op

import attr
import copy
import logging
import pytest

from niceman.cmd import GitRunner
from niceman.distributions.vcs import VCSTracer
from niceman.support.exceptions import CommandError
from niceman.utils import chpwd
from niceman.utils import swallow_logs
from niceman.tests.utils import assert_is_subset_recur
from niceman.tests.utils import create_tree
from niceman.tests.utils import with_tree
from niceman.tests.fixtures import git_repo_fixture, svn_repo_fixture
from niceman.distributions.vcs import GitDistribution
from niceman.distributions.vcs import GitRepo

git_repo_empty = git_repo_fixture(kind="empty")
git_repo = git_repo_fixture()
git_repo_pair = git_repo_fixture(kind="pair")
git_repo_pair_module = git_repo_fixture(kind="pair", scope="module")

svn_repo_empty = svn_repo_fixture(kind='empty')
svn_repo = svn_repo_fixture()


# TODO: Move to niceman.test.utils and use in other tracer tests.
def assert_distributions(result, expected_length=None, which=0,
                         expected_unknown=None, expected_subset=None):
    """Wrap common assertions about identified distributions.

    Parameters
    ----------
    result : iterable
        The result of a tracer's identify_distributions method (in its
        original generator form or as a list).
    expected_length : int, optional
        Expected number of items in `result`.
    which : int, optional
        Index specifying which distribution from result to consider
        for the `expected_unknown` and `expected_subset`
        assertions.
    expected_unknown : list, optional
        Which files should be marked as unknown in distribution
        `which`.
    expected_subset : dict, optional
        This dict is expected to be a subset of the distribution
        `which`.  The check is done by `assert_is_subset_recur`.
    """
    result = list(result)

    if expected_length is not None:
        assert len(result) == expected_length

    dist, unknown_files = result[which]

    if expected_unknown is not None:
        assert unknown_files == expected_unknown

    if expected_subset is not None:
        assert_is_subset_recur(expected_subset, attr.asdict(dist), [dict, list])


def test_git_repo_empty(git_repo_empty):
    tracer = VCSTracer()
    # Should not crash when given path to empty repo.
    assert_distributions(
        tracer.identify_distributions([git_repo_empty]),
        expected_length=1,
        expected_unknown=set(),
        expected_subset={"name": "git",
                         "packages": [{"path": git_repo_empty,
                                       "branch": "master",
                                       # We do not include repo path itself.
                                       "files": []}]})


def test_git_repo(git_repo):
    paths = [
        # Both full ...
        os.path.join(git_repo, "foo"),
        # ... and relative paths work.
        "bar",
        # So do paths in subdirectories.
        os.path.join(git_repo, "subdir/baz")
    ]

    tracer = VCSTracer()

    with chpwd(git_repo):
        dists = list(tracer.identify_distributions(paths + ["/sbin/iptables"]))
        assert_distributions(
            dists,
            expected_length=1,
            expected_unknown={"/sbin/iptables"},
            expected_subset={
                "name": "git",
                "packages": [{"files": [op.relpath(p) for p in paths],
                              "path": git_repo,
                              "branch": "master"}]})

        assert dists[0][0].packages[0].hexsha
        assert dists[0][0].packages[0].root_hexsha

        runner = GitRunner()
        hexshas, _ = runner(["git", "rev-list", "master"], cwd=git_repo)
        root_hexsha = hexshas.strip('\n').split('\n')[-1]
        repo = dists[0][0].packages[0]
        assert repo.root_hexsha == root_hexsha
        assert repo.identifier == repo.root_hexsha
        assert repo.commit == repo.hexsha

        # Above we identify a subdirectory file, but we should not
        # identify the subdirectory itself because in principle Git is
        # not tracking directories.
        subdir = os.path.join(git_repo, "subdir")
        assert not list(tracer.identify_distributions([subdir]))


def test_git_repo_detached(git_repo):
    runner = GitRunner()
    # If we are in a detached state, we still don't identify the
    # repository itself.
    runner(["git", "checkout", "master^{}", "--"],
           cwd=git_repo, expect_stderr=True)

    hexsha_master, _ = runner(["git", "rev-parse", "master"],
                              cwd=git_repo)
    hexsha_master = hexsha_master.strip()

    tracer = VCSTracer()
    dists = list(tracer.identify_distributions([git_repo]))

    pkg = dists[0][0].packages[0]
    # We do not include repository path itself.
    assert pkg.files == []
    assert pkg.hexsha == hexsha_master
    assert not pkg.branch
    assert not pkg.remotes


def test_git_repo_remotes(git_repo_pair):
    repo_local, repo_remote = git_repo_pair
    runner = GitRunner()
    tracer = VCSTracer()

    # Set remote.pushdefault to a value we know doesn't exist.
    # Otherwise, the test machine may have remote.pushdefault globally
    # configured to point to "origin".
    runner(["git", "config", "remote.pushdefault", "notexisting"],
           cwd=repo_local)
    # Add another remote that doesn't contain the current commit (in
    # fact doesn't actually exist), so that we test the "listed as
    # remote but doesn't contain" case.
    runner(["git", "remote", "add", "fakeremote", "fakepath"],
           cwd=repo_local)

    paths = [os.path.join(repo_local, "foo")]

    dists_nopush = list(tracer.identify_distributions(paths))
    assert_distributions(
        dists_nopush,
        expected_length=1,
        expected_subset={
            "name": "git",
            "packages": [
                {"files": [op.relpath(p, repo_local) for p in paths],
                 "path": repo_local,
                 "branch": "master",
                 "tracked_remote": "origin",
                 "remotes": {"origin":
                             {"url": repo_remote,
                              "contains": True},
                             "fakeremote":
                             {"url": "fakepath"}}}]})
    pkg_nopush = dists_nopush[0][0].packages[0]
    assert set(pkg_nopush.remotes.keys()) == {"origin", "fakeremote"}

    # fakeremote, which doesn't contain the current commit, doesn't
    # have contains=True.
    assert "contains" in pkg_nopush.remotes["origin"]
    assert "contains" not in pkg_nopush.remotes["fakeremote"]
    # pushurl is not included in the output above because it is not
    # set.
    assert "pushurl" not in list(pkg_nopush.remotes.values())

    # If we set the pushurl and retrace, it is included.
    runner(["git", "config", "remote.origin.pushurl", repo_remote],
           cwd=repo_local)
    dists_push = list(tracer.identify_distributions(paths))
    pkg_push = dists_push[0][0].packages[0]
    assert pkg_push.remotes["origin"]["pushurl"] == repo_remote

    # If we are at a commit that none of the remotes are known to
    # contain, there are no listed remotes.
    with chpwd(repo_local):
        runner(["git", "commit", "--allow-empty", "-m", "empty commit"])

    dists_nocontain = list(tracer.identify_distributions(paths))
    assert not dists_nocontain[0][0].packages[0].remotes.values()

    # The remote repository, however, doesn't have a remote, so there
    # are not listed remotes.
    paths = [os.path.join(repo_remote, "foo")]
    dists_remote = list(tracer.identify_distributions(paths))
    assert not dists_remote[0][0].packages[0].remotes.values()


def test_git_install_no_remote():
    dist = GitDistribution(name="git",
                           packages=[GitRepo(path="/tmp/shouldnt/matter")])

    with swallow_logs(new_level=logging.WARNING) as log:
        dist.initiate()
        dist.install_packages()
        assert "No remote known" in log.out


@with_tree(tree={"foo": ""})
def test_git_install_skip_existing_nongit(path=None):
    with swallow_logs(new_level=logging.WARNING) as log:
        dist_dir = GitDistribution(
            name="git",
            packages=[
                GitRepo(path=path,
                        remotes={"origin": {"url": "doesnt-matter",
                                            "contains": True}})])
        dist_dir.install_packages()
        assert "not a Git repository; skipping" in log.out

    with swallow_logs(new_level=logging.WARNING) as log:
        dist_dir = GitDistribution(
            name="git",
            packages=[
                GitRepo(path=op.join(path, "foo"),
                        remotes={"origin": {"url": "doesnt-matter",
                                            "contains": True}})])
        dist_dir.install_packages()
        assert "not a directory; skipping" in log.out


def test_git_install_skip_different_git(git_repo):
    with swallow_logs(new_level=logging.WARNING) as log:
        dist_dir = GitDistribution(
            name="git",
            packages=[
                GitRepo(path=git_repo,
                        root_hexsha="definitely doesn't match",
                        remotes={"origin": {"url": "doesnt-matter",
                                            "contains": True}})])
        dist_dir.install_packages()
        assert "doesn't match expected hexsha; skipping" in log.out


@pytest.fixture(scope="module")
def traced_repo(git_repo_pair_module):
    """Return a Git repo pair and the traced GitDistribution for the local repo.
    """
    repo_local, repo_remote = git_repo_pair_module

    runner = GitRunner(cwd=repo_local)
    runner(["git", "remote", "add", "dummy-remote", "nowhere"])

    tracer = VCSTracer()
    dists = list(tracer.identify_distributions([op.join(repo_local, "foo")]))
    git_dist = dists[0][0]
    assert len(git_dist.packages) == 1

    return {"repo_local": repo_local,
            "repo_remote": repo_remote,
            "git_dist": git_dist}


@pytest.fixture(scope="function")
def traced_repo_copy(traced_repo):
    return copy.deepcopy(traced_repo)


def install(git_dist, dest, check=False):
    """Helper to install GitDistribution `git_dist` to `dest`.

    If `check`, trace the installed repository and run a basic comparison in
    the GitRepo objects.
    """
    tracer = VCSTracer()
    git_dist.install_packages()

    if check:
        dists_installed = list(
            tracer.identify_distributions([op.join(dest, "foo")]))
        git_dist_installed = dists_installed[0][0]
        assert len(git_dist_installed.packages) == 1
        git_pkg = git_dist.packages[0]
        git_pkg_installed = git_dist_installed.packages[0]

        for att in ["hexsha", "root_hexsha", "tracked_remote", "remotes"]:
            assert getattr(git_pkg, att) == getattr(git_pkg_installed, att)


def current_hexsha(runner):
    return runner(["git", "rev-parse", "HEAD"])[0].strip()


def current_branch(runner):
    try:
        out = runner(["git", "symbolic-ref", "--short", "HEAD"],
                     expect_fail=True)
    except CommandError:
        return
    return out[0].strip()


@pytest.mark.integration
def test_git_install(traced_repo_copy, tmpdir):
    git_dist = traced_repo_copy["git_dist"]
    git_pkg = git_dist.packages[0]
    tmpdir = str(tmpdir)

    # Install package to a new location.
    install_dir = op.join(tmpdir, "installed")
    git_pkg.path = install_dir

    install(git_dist, install_dir, check=True)
    # Installing a second time works if the root hexsha's match.
    install(git_dist, install_dir, check=True)

    runner = GitRunner(cwd=install_dir)

    # We don't try to change the state of the repository if it's dirty.
    runner(["git", "reset", "--hard", "HEAD^"])
    hexsha_existing = current_hexsha(runner)
    create_tree(install_dir, {"dirt": "dirt"})
    with swallow_logs(new_level=logging.WARNING) as log:
        install(git_dist, install_dir)
        assert "repository is dirty" in log.out
    assert current_hexsha(runner) == hexsha_existing

    # We end up on the intended commit (detached) if the existing installation
    # repo is clean.
    os.remove(op.join(install_dir, "dirt"))
    install(git_dist, install_dir)
    assert current_hexsha(runner) == git_pkg.hexsha
    assert not current_branch(runner)


@pytest.mark.integration
def test_git_install_detached(traced_repo_copy, tmpdir):
    git_dist = traced_repo_copy["git_dist"]
    git_pkg = git_dist.packages[0]
    tmpdir = str(tmpdir)

    # Install package to a new location.
    install_dir = op.join(tmpdir, "installed")
    git_pkg.path = install_dir
    runner = GitRunner(cwd=install_dir)
    install(git_dist, install_dir)

    # We detach if there is no recorded branch.
    git_pkg.branch = None
    runner(["git", "checkout", "master"])
    install(git_dist, install_dir)
    assert current_hexsha(runner) == git_pkg.hexsha
    assert not current_branch(runner)


@pytest.mark.integration
def test_git_install_hexsha_not_found(traced_repo_copy, tmpdir):
    git_dist = traced_repo_copy["git_dist"]
    git_pkg = git_dist.packages[0]
    tmpdir = str(tmpdir)

    # Install package to a new location.
    install_dir = op.join(tmpdir, "installed")
    git_pkg.path = install_dir
    install(git_dist, install_dir)

    # No existing hexsha.
    git_pkg.hexsha = "0" * 40
    with swallow_logs(new_level=logging.WARNING) as log:
        install(git_dist, install_dir)
        assert "expected hexsha wasn't found" in log.out


@pytest.mark.integration
def test_git_install_checkout(traced_repo_copy, tmpdir):
    git_dist = traced_repo_copy["git_dist"]
    git_pkg = git_dist.packages[0]
    tmpdir = str(tmpdir)

    install_dir = op.join(tmpdir, "installed")
    runner = GitRunner(cwd=install_dir)

    # Installing to a non-existing location will be more aggressive, creating a
    # new branch at the recorded hexsha rather than just detaching there.
    git_pkg.path = install_dir
    git_pkg.branch = "other"
    install(git_dist, install_dir, check=True)
    assert current_branch(runner) == "other"

    # If the recorded branch is in the existing installation repo and has the
    # same hexsha, we check it out.
    runner(["git", "checkout", "master"])
    runner(["git", "branch", "--force", "other", git_pkg.hexsha])
    install(git_dist, install_dir)
    assert current_hexsha(runner) == git_pkg.hexsha
    assert current_branch(runner) == "other"
    # Otherwise, we detach.
    runner(["git", "checkout", "master"])
    runner(["git", "branch", "--force", "other", git_pkg.hexsha + "^"])
    install(git_dist, install_dir)
    assert current_hexsha(runner) == git_pkg.hexsha
    assert not current_branch(runner)


@pytest.mark.integration
def test_git_install_add_remotes(traced_repo_copy, tmpdir):
    git_dist = traced_repo_copy["git_dist"]
    git_pkg = git_dist.packages[0]
    tmpdir = str(tmpdir)

    install_dir = op.join(tmpdir, "installed")
    runner = GitRunner(cwd=install_dir)

    git_pkg.path = install_dir
    git_pkg.tracked_remote = "foo"
    url = git_pkg.remotes["origin"]["url"]
    git_pkg.remotes = {"foo": {"url": url},
                       "bar": {"contains": True, "url": url}}
    install(git_dist, install_dir)
    installed_remotes = runner(["git", "remote"],
                               cwd=install_dir)[0].splitlines()
    assert set(installed_remotes) == {"foo", "bar"}


def test_svn(svn_repo):
    (svn_repo_root, checked_out_dir) = svn_repo
    svn_file = os.path.join(checked_out_dir, 'foo')
    uuid_file = os.path.join(svn_repo_root, 'db', 'uuid')
    uuid = open(uuid_file).readlines()[0].strip()
    tracer = VCSTracer()
    assert_distributions(
        tracer.identify_distributions([svn_file]),
        expected_length=1,
        expected_subset={'name': 'svn'})
    svn_repo = list(tracer.identify_distributions([svn_file]))[0][0].packages[0]
    assert svn_repo.files == ['foo']
    assert svn_repo.uuid == uuid
    assert svn_repo.root_url == 'file://' + svn_repo_root
    assert svn_repo.revision == 1
    assert svn_repo.identifier == svn_repo.uuid
    assert svn_repo.commit == svn_repo.revision


def test_empty_svn(svn_repo_empty):
    (svn_repo_root, checked_out_dir) = svn_repo_empty
    tracer = VCSTracer()
    distributions = list(tracer.identify_distributions([checked_out_dir]))
    svn_repo = distributions[0][0].packages[0]
    assert svn_repo.revision is None



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from __future__ import absolute_import

import io

from niceman.distributions.base import EnvironmentSpec
from niceman.distributions.conda import CondaChannel
from niceman.distributions.conda import CondaDistribution
from niceman.distributions.conda import CondaEnvironment
from niceman.distributions.conda import CondaPackage
from niceman.distributions.debian import APTSource
from niceman.distributions.debian import DEBPackage
from niceman.distributions.debian import DebianDistribution
from niceman.distributions.vcs import GitDistribution
from niceman.distributions.vcs import GitRepo
from niceman.distributions.vcs import SVNDistribution
from niceman.distributions.vcs import SVNRepo
from niceman.distributions.venv import VenvDistribution
from niceman.distributions.venv import VenvEnvironment
from niceman.distributions.venv import VenvPackage
from niceman.formats.niceman import NicemanProvenance

from .constants import NICEMAN_SPEC1_YML_FILENAME


def test_write():
    output = io.StringIO()
    # just load
    file_format = NicemanProvenance(NICEMAN_SPEC1_YML_FILENAME)
    env = file_format.get_environment()
    # just a basic test that we loaded stuff correctly
    assert len(env.distributions) == 2
    assert env.distributions[0].name == 'conda'
    assert len(env.distributions[1].apt_sources) == 3
    # and save
    NicemanProvenance.write(output, env)
    out = output.getvalue()
    env_reparsed = NicemanProvenance(out).get_environment()
    # and we could do the full round trip while retaining the same "value"
    assert env == env_reparsed


def test_spec_round_trip():
    # FIXME: This should also test GitDistribution's, but NicemanProvenance
    # can't currently load those (gh-222).

    spec = EnvironmentSpec(
        distributions=[
            DebianDistribution(
                name="debian",
                apt_sources=[
                    APTSource(
                        name="apt_Debian_stable_main_0",
                        component="main",
                        archive="stable",
                        architecture="amd64",
                        codename="stretch",
                        origin="Debian",
                        label="Debian",
                        site="ftp.us.debian.org",
                        archive_uri="http://ftp.us.debian.org/debian",
                        date="2018-03-10 10:21:19+00:00")],
                packages=[
                    DEBPackage(name="debpackage"),
                    DEBPackage(
                        name="libc-bin",
                        upstream_name=None,
                        version="2.24-11+deb9u3",
                        architecture="amd64",
                        source_name="glibc",
                        source_version=None,
                        size="779468",
                        md5="3b9aaa83b5253895b8e13509659662e4",
                        sha1=None,
                        sha256="aaa",
                        versions={"2.24-11+deb9u1": ["apt_Debian_stable_foo"],
                                  "2.24-11+deb9u3": ["apt_Debian_stable_bar",
                                                     "apt__now__0"]},
                        install_date="2018-03-12 10:55:13+00:00",
                        files=["/usr/bin/zdump"])],
                version="9.4"),
            CondaDistribution(
                name="conda",
                path="/path/to/miniconda3",
                conda_version="4.4.10",
                python_version="3.6.3.final.0",
                platform="linux-64",
                environments=[
                    CondaEnvironment(
                        name="root",
                        path="/path/to/miniconda3",
                        packages=[
                            CondaPackage(name="condapkg"),
                            CondaPackage(
                                version="36.5.0",
                                build="py36he42e2e1_0",
                                name="setuptools",
                                md5="cb1383539629db998105faf7e91e2bc7",
                                url="https://somewhere")],
                        channels=[
                            CondaChannel(
                                name="defaults",
                                url="https://somewhere")]),
                    CondaEnvironment(
                        name="other",
                        path="/path/to/miniconda3",
                        packages=[
                            CondaPackage(name="condapkg2")])]),
            GitDistribution(
                name="git",
                packages=[GitRepo(path="/path/to/repo")]),
            SVNDistribution(
                name="svn",
                packages=[SVNRepo(path="/path/to/repo")]),
            VenvDistribution(
                name="venv0",
                path="/usr/bin/virtualenv",
                venv_version="15.1.0",
                environments=[
                    VenvEnvironment(
                        path="venv-niceman",
                        python_version="3.5.3",
                        packages=[
                            VenvPackage(
                                version="3.12",
                                name="PyYAML",
                                location="/path/to/venv/site-packages",
                                local=True)])]),
            VenvDistribution(name="venv1")])

    output = io.StringIO()
    NicemanProvenance.write(output, spec)
    loaded = NicemanProvenance(output.getvalue()).get_environment()
    assert spec == loaded



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import logging

from niceman.utils import swallow_logs
from niceman.formats import Provenance


def test_get_distributions(demo1_spec):

    # Test reading the distributions from the NICEMAN spec file.
    provenance = Provenance.factory(demo1_spec, 'niceman')

    with swallow_logs(new_level=logging.DEBUG) as log:
        distributions = provenance.get_distributions()
        assert len(distributions) == 2
        # a bit of testing is done within test_niceman.py since it is niceman specific example?



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Generic interface(s) for handling session interactions."""

import logging

lgr = logging.getLogger('niceman.resource.session')

import attr
from functools import partial
import os
import os.path as op
import re
from six.moves import shlex_quote

from niceman.support.exceptions import SessionRuntimeError
from niceman.cmd import Runner
from niceman.dochelpers import exc_str, borrowdoc
from niceman.support.exceptions import CommandError
from niceman.utils import updated
from niceman.utils import to_unicode

import logging
lgr = logging.getLogger('niceman.session')


@attr.s
class Session(object):
    """Interface for Resources to provide interaction within that environment"""

    INTERNAL_COMMANDS = ['mkdir', 'isdir', 'put', 'get', 'chown', 'chmod']


    def __attrs_post_init__(self):
        """
        Maintain both current and future session environments.

        For persistent resources, we will save the environment information
        to make it available for sessions beyond the current one.
        """
        self._env = {}           # environment which would be in-effect only for this session
        self._env_permanent = {} # environment variables which would be in-effect in future sessions if resource is persistent

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __call__(self, *args, **kwargs):
        """Sugaring shortcut to `execute_command`"""
        return self.execute_command(*args, **kwargs)

    # By default don't do anything special
    def open(self):
        """
        Called when a session is started.
        """
        # XXX
        # we might right here already load possibly set permanent env variables
        # within that session?
        pass

    def close(self):
        """
        Called when a session ends.
        """
        # XXX may be here we should dump permanent env settings?
        pass

    def set_envvar(self, variable, value=None, permanent=False, format=False):
        """Set environment variable(s) to be used within the session

        All of them would be exported (TODO: should we allow for non-exported
        ones for some reason?  we aren't really scripting to care about non-exported)

        Parameters
        ----------
        variable: str or dict
          Variable name or entire dictionary of variable: value pairs (value
          must be undefined then).
        value: str or None
          If variable is not a dict, it would be the value for the variable
          to assign.  If value is None, removes that variable. Use empty string
          if you want it to be defined and empty
        permanent: bool, optional
          Store that variable for future session within this "container"
        format: bool
          Use Python string `.format()` giving old value of the env dictionary
        """
        # XXX this is a generic handling but may be we should allow some
        # backends to handle it internally?  e.g. docker's ENV vs ARG
        for env in [self._env] + ([self._env_permanent] if permanent else []):
            if isinstance(variable, dict):
                assert value is None, "Must not provide new env as dict and a value"
                update_env = variable
            else:
                update_env = {variable: value}

            for newvar, newvalue in update_env.items():
                if newvalue is None and newvar in env:
                    del env[newvar]
                else:
                    if format:
                        newvalue = newvalue.format(env[newvar])
                    env[newvar] = newvalue
        if permanent:
            # We should store adjusted environment within the session for future
            # invocation
            raise NotImplementedError

    def get_envvars(self, permanent=False):
        """Get stored session environment variables

        Parameters
        ----------
        permanent : {bool}, optional
            Indicate that the environment variables that persist across
            sessions should be returned (the default is False, which will
            return the environment variables specific to the current session.)

        Returns
        -------
        dict
            Key/value pairs of environment variables
        """
        # TODO: should we parametrize to be able to query for permanent ones
        # if were defined or those we store in our variables etc?
        return self._env_permanent if permanent else self._env

    def query_envvars(self):
        """Query full session environment settings within the session"""
        raise NotImplementedError

    def source_script(self, command, permanent=False, diff=True, shell=None):
        """Source a script which would modify the environment

        Note: if command is composite (e.g. "activate envname" for conda), it
        would work only in bash or zsh shell.

        Parameters
        ----------
        command : str or list
          Name of the script or composite command (if a list, such as
          ["activate", "envname"] in conda) to be "sourced"
        permanent : bool, optional
        diff : bool, optional
          Store only variables values of which were changed by sourcing the file
        shell : str, optional
          Which shell to use.  If none specified, the one specified by SHELL
          in the environment would be used. If that one is not specified -- /bin/sh
          will be used for simple command, or /bin/bash if composite

        Returns
        -------
        dict
            Key/value pairs of the new environment
        """
        raise NotImplementedError

    # TODO: move logic/handling of batched commands defined in
    # Resource  and probably env vars handling

    def execute_command(self, command, env=None, cwd=None):
        """
        Execute the given command in the environment.

        Parameters
        ----------
        command : list or str
            Shell command string or list of command tokens to send to the
            environment to execute.
        env : dict, optional
            Additional environment variables which are applied
            only to the current call.  If value is None -- variable will be
            removed
        cwd : string, optional
            Path of directory in which to run the command

        Returns
        -------
        (stdout, stderr)
        """
        # TODO: bring back updated_env?
        command_env = dict(self._env, **(env or {}))
        run_kw = {}
        if command_env:
            run_kw['env'] = command_env

        return self._execute_command(
            command,
            cwd=cwd,
            **run_kw
        )  # , shell=True)

    def _execute_command(self, command, env=None, cwd=None):
        """
        Execute the given command in the environment.

        Parameters
        ----------
        command : list or str
            Shell command string or list of command tokens to send to the
            environment to execute.
        env : dict, optional
            Additional environment variables which are applied
            only to the current call.  If value is None -- variable will be
            removed
        cwd : string, optional
            Path of directory in which to run the command

        Returns
        -------
        (stdout, stderr)

        Raises
        ------
        CommandError
           if command's exitcode wasn't 0 or None. exitcode is passed to
           CommandError's `code`-field. Command's stdout and stderr are stored
           in CommandError's `stdout` and `stderr` fields respectively.
        """
        raise NotImplementedError

    #
    # Files query and manipulation
    # TODO:  should be in subspace (.path) may be? This would allow for
    #        more flexible mixups

    def niceman_exec(self, command, args):
        """Run a niceman utility "exec" command in the environment

        Parameters
        ----------
        command : string
            The session method name to run. (e.g. mkdir, chown, etc.)
        args : list of strings
            Arguments passed in from the command line for the command

        Raises
        ------
        CommandError
            Exception if an invalid command argument is passed in.
        """

        if command not in self.INTERNAL_COMMANDS:
            raise CommandError(cmd=command, msg="Invalid command")

        pargs = [] # positional args to pass to session command
        kwargs = {} # key word args to pass to session command
        for arg in args:
            if '=' in arg:
                parts = arg.split('=')
                if len(parts) != 2:
                    raise CommandError(cmd=command,
                                       msg="Invalid command line parameter")
                kwargs[parts[0]] = parts[1]
            else:
                pargs.append(arg)

        getattr(self, command)(*pargs, **kwargs)

    def exists(self, path):
        """Determine if a path exists in the resource.

        Parameters
        ----------
        path : string
            Path to check for existence in the resource environment.

        Returns
        -------
        bool
            True if path exists
        """
        raise NotImplementedError

    def _prepare_dest_path(self, src_path, dest_path,
                           local=True, absolute_only=False):
        """Do common handling for the destination target of `get` and `put`.

        Parameters
        ----------
        src_path : str
            Path to source file.
        dest_path : str
            Path to target file.  If `dest_path` is None, the basename is taken
            from `src_path`.
        local : bool, optional
            Whether the destination is on the local machine.
        absolute_only : bool, optional
            Whether `dest_path` is required to be absolute.

        Returns
        -------
        The path to the destination, possibly adjusted to add the basename.
        """
        if local:
            exists = op.exists
            mkdir = os.makedirs
        else:
            exists = self.exists
            mkdir = partial(self.mkdir, parents=True)

        if absolute_only and not op.isabs(dest_path):
            raise ValueError(
                "Destination path must be absolute, got {}".format(dest_path))

        dest_dir = dest_base = None
        if dest_path:
            dest_dir, dest_base = op.split(dest_path)
            if dest_dir and not exists(dest_dir):
                mkdir(dest_dir)

        if not dest_base:
            dest_base = op.basename(src_path)
            dest_path = op.join(dest_dir, dest_base) if dest_dir else dest_base
        return dest_path

    def put(self, src_path, dest_path, uid=-1, gid=-1):
        """Take file on the local file system and copy over into the resource

        The src_path and dest_path must include the name of the file being
        transferred.

        Parameters
        ----------
        src_path : string
            Path to file to push to resource environment
        dest_path : string
            Path of resource directory to put local file in.  If this contains
            a trailing separator, it is considered directory and the base name
            is taken from `src_path`.
        uid : int, optional
            System user ID to assign ownership of file on resource  (the
            default is -1, which will preserve the user owner of the local file)
        gid : int, optional
            System group ID to assign group ownership of file on resource (the
            default is -1, which will preserve the group id of the local file)
        """
        raise NotImplementedError

    def get(self, src_path, dest_path=None, uid=-1, gid=-1):
        """Take file on the resource and copy over into the local system

        The src_path must include the name of the file being transferred.

        Parameters
        ----------
        src_path : string
            Path to file to pull from resource environment
        dest_path : string, optional
            Path in local file system to put local file in.  If unspecified,
            the destination path defaults the base name of `src_path` within
            the current directory.
        uid : int, optional
            System user ID to assign ownership of file on resource  (the
            default is -1, which will preserve the user owner of the local file)
        gid : int, optional
            System group ID to assign group ownership of file on resource (the
            default is -1, which will preserve the group id of the local file)
        """
        raise NotImplementedError

    def get_mtime(self, path):
        """Returns the modification time for a file in the resource

        Parameters
        ----------
        path : string
            Path to file on resource

        Returns
        -------
        string
            epoch timestamp

        """
        raise NotImplementedError

    #
    # Somewhat optional since could be implemented with native "POSIX" commands
    #
    def read(self, path, mode='r'):
        """Return content of a file

        Parameters
        ----------
        path : string
            Filesystem path to file to be read
        mode : string, optional
            Access mode when reading file (the default is 'r', which is read-only)

        Returns
        -------
        list of strings
            Contents of the file, one output line for each list item
        """
        raise NotImplementedError

    def mkdir(self, path, parents=False):
        """Create a directory

        Parameters
        ----------
        path : string
            Path to file in resource
        parents : bool, optional
            Include parent directories (the default is False)
        """
        raise NotImplementedError

    def mktmpdir(self):
        """Create a temporary directory and return the path

        Returns
        -------
        path : String
            Path to the created temporary directory
        """
        raise NotImplementedError

    def isdir(self, path):
        """Return True if path is pointing to a directory

        Parameters
        ----------
        path : string
            Path to test in the resource

        Returns
        -------
        bool
            True if path is a directory
        """
        raise NotImplementedError

    def chmod(self, path, mode, recursive=False):
        """Set the mode of the indicated path

        Parameters
        ----------
        path : string
            Path to file or directory whose permissions we want to update
        mode : string
            Octal number representing the permission bit pattern
        recursive : bool, optional
            Recurse the permission updates into the subdirectories (the default
            is False)
        """
        raise NotImplementedError

    def chown(self, path, uid=-1, gid=-1, recursive=False, remote=True):
        """Set the owner and group ownership of a file or directory.

        Parameters
        ----------
        path : string
            Path to file or directory whose ownship we want to update
        uid : number, optional
            System id of user (the default is -1, which leaves the existing
            user owner as is)
        gid : number, optional
            System id of group (the default is -1, which leaves the existing
            group owner as is)
        recursive : bool, optional
            Recurse into the path directory provided and update the ownership
            to all files and directories. (the default is False, which does
            not recurse)
        remote : bool, optional
            Runs the chown command to the path in the resource environment (the
            default is True, which runs in the resource environment, False
            runs the chown command in the local, host file system)
        """
        raise NotImplementedError


@attr.s
class POSIXSession(Session):
    """A Session which relies on commands present in any POSIX-compliant env

    """

    # -0 is not provided by busybox's env command.  So if we decide to make it
    # even more portable - something to be done
    _GET_ENVIRON_CMD = ['env', '-0']

    @borrowdoc(Session)
    def query_envvars(self):
        out, err = self.execute_command(self._GET_ENVIRON_CMD)
        env = self._parse_envvars_output(out)
        # TODO:  should we update with our .env or .env_permament????
        return env

    def _parse_envvars_output(self, out):
        """Decode a JSON string into an object

        Parameters
        ----------
        out : string
            JSON string to decode.

        Returns
        -------
        object
            Decoded representation of the JSON string
        """
        output = {}
        for line in to_unicode(out).split('\0'):
            if not line:
                continue
            split = line.split('=', 1)
            if len(split) != 2:
                lgr.warning(
                    "Failed to split envvar definition into key=value. Got %s", line)
                continue
            output[split[0]] = split[1]
        return output

    @borrowdoc(Session)
    def source_script(self, command, permanent=False, diff=True, shell=None):
        orig_env = self.query_envvars()
        # Might want to be reimplemented in derived classes?  e.g.
        # if session is persistent (i.e all commands run in persistent session)
        # and we don't need this source  to be permanent -- we could
        # just run it and be done
        marker = "== =NICEMAN == ="  # unique marker to be able to split away
        # possible output from the sourced script
        get_env_command = " ".join('"%s"' % s for s in self._GET_ENVIRON_CMD)
        shell = shell or self.query_envvars().get('SHELL', None)
        if not isinstance(command, list):
            command = [command]
            shell = shell or "/bin/sh"
        else:
            # apparently in purely POSIX shell (e.g. dash) you cannot do
            # parametric "source" calls, e.g. ". activate datalad".
            # Moreover e.g. conda supports only bash and zsh
            shell = shell or "/bin/bash"
        command = " ".join('"%s"' % c for c in command)
        out, err = self.execute_command(
            # is a composite command which would first source
            # and then run the same command we use in self.query_envvars
            # if it is all permanent
            [
                # Possibly use SHELL which was already set within the environment
                shell,
                '-c',
                '. {command}; echo "{marker}"; {get_env_command}'.format(**locals())
            ]
        )
        # stderr is ok -- above call might issue a warning
        # assert not err
        new_env = self._parse_envvars_output(
            re.sub('.*%s\n*' % marker, '', out, flags=re.DOTALL)
        )

        # TODO: deal with possible removals, so we should prune them from
        # local envs as well, warning if some variable wasn't even explicitly
        # listed locally

        # Apparently whenever it is a "parametric" command then shells insert
        # some "goodness" which I think we shouldn't care about
        if shell in ["zsh", "/bin/zsh"]:
            new_env.pop('_')
            new_env.pop('OLDPWD')
        elif shell in ["bash", "/bin/bash"]:
            new_env.pop('_')
            new_env.pop('SHLVL')

        if diff:
            # minimize by dropping the same as before
            for k in list(new_env.keys()):
                if k in orig_env and orig_env[k] == new_env[k]:
                    new_env.pop(k)

        env = self._env_permanent if permanent else self._env
        for k, v in new_env.items():
            env[k] = v

        return new_env

    def exists(self, path):
        """Return if file exists"""
        try:
            out, err = self.execute_command(self.exists_command(path))
        except Exception as exc:  # TODO: More specific exception?
            lgr.debug("Check for file presence failed: %s", exc_str(exc))
            return False
        if out == 'Found\n':
            return True
        else:
            lgr.debug("Standard error was not empty (%r), thus assuming that "
                      "test for file presence has failed", err)
            return False

    def exists_command(self, path):
        """Return the command to run for the exists method."""
        command = ['test', '-e', shlex_quote(path), '&&', 'echo', 'Found']
        return ['bash', '-c', ' '.join(command)]

    # def lexists(self, path):
    #     """Return if file (or just a broken symlink) exists"""
    #     return os.path.lexists(path)

    # Seems to have no generic implementation in POSIX?  TODO: check
    #  may be we could assume presence of e.g. python so we could use std library?
    def get_mtime(self, path):
        # TODO:  too common of a pattern -- we need a helper to wrap such calls
        out, err = self.execute_command(self.get_mtime_command(path))
        return out.strip()

    def get_mtime_command(self, path):
        """Return the command to run for the get_mtime method."""
        command = "import os, sys; print(os.path.getmtime(sys.argv[1]))"
        return ['python', '-c', command, path]

    #
    # Somewhat optional since could be implemented with native "POSIX" commands
    #
    def read(self, path, mode='r'):
        """Return context manager to open files for reading or editing"""
        out, err = self.execute_command(["cat", path])
        if err:
            raise SessionRuntimeError("Running had std error output: %s" % err)
        return out

    def mkdir(self, path, parents=False):
        """Create a directory
        """
        command = ["mkdir"]
        if parents: command.append("-p")
        command += [path]
        self.execute_command(command)

        if not self.isdir(path):
            raise CommandError(cmd='mkdir', msg="Failed to create directory")

    def mktmpdir(self):
        path, _ = self.execute_command(["mktemp", "-d"])
        return path.rstrip()  # Remove newline

    def isdir(self, path):
        try:
            out, err = self.execute_command(self.isdir_command(path))
        except Exception as exc:  # TODO: More specific exception?
            lgr.debug("Check for directory failed: %s", exc_str(exc))
            return False
        if out == 'Found\n':
            return True
        else:
            lgr.debug("Standard error was not empty (%r), thus assuming that "
                      "test for direcory has failed", err)
            return False

    def isdir_command(self, path):
        """Return the command to run for the isdir method."""
        command = ['test', '-d', shlex_quote(path), '&&', 'echo', 'Found']
        return ['bash', '-c', ' '.join(command)]

    def chmod(self, path, mode, recursive=False):
        """Set the mode of a remote path
        """
        command = ['chmod']
        if recursive: command += ["-R"]
        command += [mode] + [path]
        self.execute_command(command)

    def chown(self, path, uid=-1, gid=-1, recursive=False, remote=True):
        """Set the user and gid of a path
        """
        uid = int(uid) # Command line parameters getting passed as type str
        gid = int(gid)

        if uid == -1 and gid > -1:
            command = ['chgrp']
        else:
            command = ['chown']
        if recursive: command += ["-R"]
        if uid > -1 and gid > -1: command += ["{}.{}".format(uid, gid)]
        elif uid > -1: command += [uid]
        elif gid > -1: command += [gid]
        else: raise CommandError(cmd='chown', msg="Invalid command \
            parameters.")
        command += [path]
        if remote:
            self.execute_command(command)
        else:
            # Run on the local file system
            Runner().run(command)


def get_local_session(env={'LC_ALL': 'C'}, pty=False, shared=None):
    """A shortcut to get a local session"""
    # TODO: support arbitrary session as obtained from a resource
    # TODO:  Shell needs a name -- should we request from manager
    #        which would assume some magical name for reuse??
    from niceman.resource.shell import Shell
    session = Shell("localshell").get_session(pty=pty, shared=shared)
    # or we shouldn't set it ? XXXX
    if env:
        session.set_envvar(env)
    return session

def get_updated_env(env, update):
    """Given an environment and set of updates, return updated one

    Special handling -- keys with None for the value, will be removed
    """
    env_ = updated(env, update)
    # pop those explicitly set to None
    for e in list(env_):
        if env_[e] is None:
            del env_[e]
    return env_



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##


from collections import OrderedDict
import logging
import pytest
from mock import patch, call, MagicMock

from niceman.api import create
from niceman.cmdline.main import main
from niceman.utils import swallow_logs
from niceman.resource.base import ResourceManager
from niceman.tests.utils import assert_in
from niceman.support.exceptions import ResourceError

from ..create import parse_backend_parameters


def test_create_interface():
    """
    Test creating an environment
    """

    with patch('docker.Client') as client, \
        patch('niceman.resource.ResourceManager._save'), \
        patch('niceman.resource.ResourceManager._get_inventory'), \
        swallow_logs(new_level=logging.DEBUG) as log:

        client.return_value = MagicMock(
            containers=lambda all: [],
            pull=lambda repository, stream: [
                b'{ "status" : "status 1", "progress" : "progress 1" }',
                b'{ "status" : "status 2", "progress" : "progress 2" }'
            ],
            create_container=lambda name, image, stdin_open, tty, command: {
                'Id': '18b31b30e3a5'
            }
        )

        args = ['create',
                '--resource-type', 'docker-container',
                '--backend', 'engine_url=tcp://127.0.0.1:2376',
                '--',
                'my-test-resource'
        ]
        with patch("niceman.interface.create.get_manager",
                   return_value=ResourceManager()):
            main(args)

        calls = [
            call(base_url='tcp://127.0.0.1:2376'),
            call().start(container='18b31b30e3a5')
        ]
        client.assert_has_calls(calls, any_order=True)

        assert_in("status 1 progress 1", log.lines)
        assert_in("status 2 progress 2", log.lines)
        assert_in("Created the environment my-test-resource", log.lines)


def test_create_missing_required():
    with pytest.raises(ResourceError) as exc:
        # SSH requires host.
        with patch("niceman.interface.create.get_manager",
                   return_value=ResourceManager()):
            create("somessh", "ssh", [])
    assert "host" in str(exc.value)


def test_parse_backend_parameters():
    for value, expected in [(["a=b"], {"a": "b"}),
                            (["a="], {"a": ""}),
                            (["a=c=d"], {"a": "c=d"}),
                            (["a-b=c d"], {"a-b": "c d"}),
                            ({"a": "c=d"}, {"a": "c=d"})]:
        assert parse_backend_parameters(value) == expected

    # We leave any mapping be, including not converting an empty mapping to an
    # empty dict.
    assert isinstance(parse_backend_parameters(OrderedDict({})),
                      OrderedDict)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Facility for managing compute resources.

"""

__docformat__ = 'restructuredtext'

from .base import Resource, ResourceManager

_MANAGER = None


def get_manager():
    """Return a ResourceManager instance.

    A new instance is not created if one already exists.  This getter function
    is used rather than a module-level instance to support delaying instance
    creation.
    """
    global _MANAGER
    if _MANAGER is None:
        _MANAGER = ResourceManager()
    return _MANAGER



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Environment sub-class to provide management of environment engine."""

import attr
import docker
import dockerpty
import io
import json
import os
import requests
import tarfile
from niceman import utils
from ..cmd import Runner
from ..support.exceptions import CommandError, ResourceError
from niceman.dochelpers import borrowdoc
from niceman.resource.session import POSIXSession, Session
from .base import Resource
from ..utils import attrib

import logging
lgr = logging.getLogger('niceman.resource.docker_container')


@attr.s
class DockerContainer(Resource):
    """
    Environment manager which manages a Docker container.
    """

    # Generic properties of any Resource
    name = attrib(default=attr.NOTHING)

    # Container properties
    id = attrib()
    type = attrib(default='docker-container')

    image = attrib(default='ubuntu:latest',
        doc="Docker base image ID from which to create the running instance")
    engine_url = attrib(default='unix:///var/run/docker.sock',
        doc="Docker server URL where engine is listening for connections")
    seccomp_unconfined = attrib(default=False,
        doc="Disable kernel secure computing mode when creating the container")

    status = attrib()

    # Docker client and container objects.
    _client = attrib()
    _container = attrib()

    @staticmethod
    def is_engine_running(base_url=None):
        """Check the local environment to see if the Docker engine is running.

        Parameters
        ----------
        base_url : str
            URL or socket where Docker engine is listening

        Returns
        -------
        boolean
        """
        try:
            session = docker.Client(base_url=base_url)
            session.info()
        except (requests.exceptions.ConnectionError,
                docker.errors.DockerException):
            return False
        return True

    @staticmethod
    def is_container_running(container_name):
        """Ping the local environment to see if given container is running.

        Parameters
        ----------
        container_name : string

        Returns
        -------
        boolean
        """
        stdout, _ = Runner().run(['docker', 'ps', '--quiet', '--filter',
            'name=^/{}$'.format(container_name)])
        if stdout.strip():
            return True
        return False

    def connect(self):
        """
        Open a connection to the environment.
        """
        # Open a client connection to the Docker engine.
        self._client = docker.Client(base_url=self.engine_url)

        containers = []
        for container in self._client.containers(all=True):
            assert self.id or self.name,\
                "Container name or id must be known"
            if self.id and not container.get('Id').startswith(self.id):
                lgr.log(5, "Container %s does not match by id: %s", container,
                        self.id)
                continue
            if self.name and ('/' + self.name) \
                    not in container.get('Names'):
                lgr.log(5, "Container %s does not match by name: %s", container,
                        self.name)
                continue
            # TODO: make above more robust and centralize across different resources/backends?
            containers.append(container)
        if len(containers) == 1:
            self._container = containers[0]
            self.id = self._container.get('Id')
            self.status = self._container.get('State')
        elif len(containers) > 1:
            raise ResourceError(
                "Multiple container matches found: %s" % str(containers)
            )
        else:
            self.id = None
            self.status = None

    def create(self):
        """
        Create a baseline Docker image and run it to create the container.

        Returns
        -------
        dict : config parameters to capture in the inventory file
        """
        if self._container:
            raise ResourceError(
                "Container '{}' (ID {}) already exists in Docker".format(
                    self.name, self.id))
        # image might be of the form repository:tag -- pull would split them
        # if needed
        for line in self._client.pull(repository=self.image, stream=True):
            status = json.loads(utils.to_unicode(line, "utf-8"))
            output = status['status']
            if 'progress' in status:
                output += ' ' + status['progress']
            lgr.info(output)
        args = {
            'name': self.name,
            'image': self.image,
            'stdin_open': True,
            'tty': True,
            'command': '/bin/bash'
        }
        # When running the rztracer binary in a Docker container, it is
        # necessary to suspend the kernel's security facility when creating
        # the container. Since it is a security issue, the default is to
        # *not* turn it off.
        if self.seccomp_unconfined:
            args['host_config'] = {
                'SecurityOpt': ['seccomp:unconfined']
            }
        self._container = self._client.create_container(**args)
        self.id = self._container.get('Id')
        self._client.start(container=self.id)
        self.status = 'running'
        return {
            'id': self.id,
            'status': self.status
        }

    def delete(self):
        """
        Deletes a container from the Docker engine.
        """
        if self._container:
            self._client.remove_container(self._container, force=True)

    def start(self):
        """
        Starts a container in the Docker engine.
        """
        if self._container:
            self._client.start(container=self._container.get('Id'))

    def stop(self):
        """
        Stops a container in the Docker engine.
        """
        if self._container:
            self._client.stop(container=self._container.get('Id'))

    def get_session(self, pty=False, shared=None):
        """
        Log into a container and get the command line
        """
        if not self._container:
            self.connect()

        if pty and shared is not None and not shared:
            lgr.warning("Cannot do non-shared pty session for docker yet")
        return (PTYDockerSession if pty else DockerSession)(
            client=self._client,
            container=self._container
        )


@attr.s
class DockerSession(POSIXSession):
    client = attrib(default=attr.NOTHING)
    container = attrib(default=attr.NOTHING)

    @borrowdoc(Session)
    def _execute_command(self, command, env=None, cwd=None):
        #command_env = self.get_updated_env(env)
        if env:
            raise NotImplementedError("passing env variables to docker session execution")

        if cwd:
            # TODO: implement
            # raise NotImplementedError("handle cwd for docker")
            lgr.warning("cwd is not handled in docker yet")
            pass
        # if command_env:
            # TODO: might not work - not tested it
            # command = ['export %s=%s' % k for k in command_env.items()] + command

        # The following call may throw the following exception:
        #    docker.errors.APIError - If the server returns an error.
        lgr.debug('Running command %r', command)
        out = []
        execute = self.client.exec_create(container=self.container, cmd=command)
        out = ''
        for i, line in enumerate(
                self.client.exec_start(exec_id=execute['Id'],
                stream=True)
        ):
            if line.startswith(b'rpc error'):
                raise CommandError(cmd=command, msg="Docker error - %s" % line)
            out += utils.to_unicode(line, "utf-8")
            lgr.debug("exec#%i: %s", i, line.rstrip())

        exit_code = self.client.exec_inspect(execute['Id'])['ExitCode']
        if exit_code not in [0, None]:
            msg = "Failed to run %r. Exit code=%d. out=%s err=%s" \
                % (command, exit_code, out, out)
            raise CommandError(str(command), msg, exit_code, '', out)
        else:
            lgr.log(8, "Finished running %r with status %s", command,
                exit_code)

        return (out, '')

    # XXX should we start/stop on open/close or just assume that it is running already?


    @borrowdoc(Session)
    def put(self, src_path, dest_path, uid=-1, gid=-1):
        # To copy one or more files to the container, the API recommends
        # to do so with a tar archive. http://docker-py.readthedocs.io/en/1.5.0/api/#copy
        dest_path = self._prepare_dest_path(src_path, dest_path,
                                            local=False, absolute_only=True)
        dest_dir, dest_basename = os.path.split(dest_path)
        tar_stream = io.BytesIO()
        tar_file = tarfile.TarFile(fileobj=tar_stream, mode='w')
        tar_file.add(src_path, arcname=dest_basename)
        tar_file.close()
        tar_stream.seek(0)
        self.client.put_archive(container=self.container['Id'], path=dest_dir,
            data=tar_stream)

        if uid > -1 or gid > -1:
            self.chown(dest_path, uid, gid)

    @borrowdoc(Session)
    def get(self, src_path, dest_path=None, uid=-1, gid=-1):
        src_dir, src_basename = os.path.split(src_path)
        dest_path = self._prepare_dest_path(src_path, dest_path)
        dest_dir = os.path.dirname(dest_path)
        stream, stat = self.client.get_archive(self.container, src_path)
        tarball = tarfile.open(fileobj=io.BytesIO(stream.read()))
        tarball.extractall(path=dest_dir)
        os.rename(os.path.join(dest_dir, src_basename), dest_path)

        if uid > -1 or gid > -1:
            self.chown(dest_path, uid, gid, remote=False)


@attr.s
class PTYDockerSession(DockerSession):
    """Interactive Docker Session"""

    @borrowdoc(Session)
    def open(self):
        lgr.debug("Opening TTY connection to docker container.")
        # TODO: probably call to super to assure that we have it running?
        dockerpty.start(self.client, self.container, logs=0)

    @borrowdoc(Session)
    def close(self):
        # XXX ?
        pass

    # XXX should we overload execute_command?


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

import logging
lgr = logging.getLogger('niceman.cmdline')

lgr.log(5, "Importing cmdline.main")

import argparse
import sys
import textwrap
from importlib import import_module

import niceman

from niceman.cmdline import helpers
from niceman.support.exceptions import InsufficientArgumentsError, MissingConfigFileError
from ..utils import setup_exceptionhook, chpwd
from ..dochelpers import exc_str


def _license_info():
    return """\
Copyright (c) 2016- NICEMAN developers (parts 2013-2016 DataLad developers)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


# TODO:  OPT look into making setup_parser smarter to become faster
# Now it seems to take up to 200ms to do all the parser setup
# even though it might not be necessary to know about all the commands etc.
# I wondered if it could somehow decide on what commands to worry about etc
# by going through sys.args first
def setup_parser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        return_subparsers=False):

    lgr.log(5, "Starting to setup_parser")
    # delay since it can be a heavy import
    from ..interface.base import dedent_docstring, get_interface_groups, \
        get_cmdline_command_name, alter_interface_docs_for_cmdline
    # setup cmdline args parser
    parts = {}
    # main parser
    parser = argparse.ArgumentParser(
        fromfile_prefix_chars='@',
        description=dedent_docstring("""\
            NICEMAN aims to ease construction and execution of computation environments
            based on collected provenance data."""),
        epilog='"Nicely Manage Your Environments"',
        formatter_class=formatter_class,
        add_help=False)
    # common options
    helpers.parser_add_common_opt(parser, 'help')
    helpers.parser_add_common_opt(parser, 'log_level')
    helpers.parser_add_common_opt(
        parser,
        'version',
        version='niceman %s\n\n%s' % (niceman.__version__, _license_info()))
    if __debug__:
        parser.add_argument(
            '--dbg', action='store_true', dest='common_debug',
            help="enter Python debugger when uncaught exception happens")
        parser.add_argument(
            '--idbg', action='store_true', dest='common_idebug',
            help="enter IPython debugger when uncaught exception happens")
    parser.add_argument(
        '-C', action='append', dest='change_path', metavar='PATH',
        help="""run as if niceman were started in <path> instead
        of the current working directory.  When multiple -C options are given,
        each subsequent non-absolute -C <path> is interpreted relative to the
        preceding -C <path>.  This option affects the interpretations of the
        path names in that they are made relative to the working directory
        caused by the -C option""")

    parser.add_argument(
        "-c", "--config", metavar="CONFIG", action="append",
        help="""path to NICEMAN configuration file.  This option can be given
        multiple times, in which case values in the later files override
        previous ones.""")

    # yoh: atm we only dump to console.  Might adopt the same separation later on
    #      and for consistency will call it --verbose-level as well for now
    # log-level is set via common_opts ATM
    # parser.add_argument('--log-level',
    #                     choices=('critical', 'error', 'warning', 'info', 'debug'),
    #                     dest='common_log_level',
    #                     help="""level of verbosity in log files. By default
    #                          everything, including debug messages is logged.""")
    #parser.add_argument('-l', '--verbose-level',
    #                    choices=('critical', 'error', 'warning', 'info', 'debug'),
    #                    dest='common_verbose_level',
    #                    help="""level of verbosity of console output. By default
    #                         only warnings and errors are printed.""")

    # subparsers
    subparsers = parser.add_subparsers()

    # auto detect all available interfaces and generate a function-based
    # API from them
    grp_short_descriptions = []
    interface_groups = get_interface_groups()
    for grp_name, grp_descr, _interfaces in interface_groups:
        # for all subcommand modules it can find
        cmd_short_descriptions = []

        for _intfspec in _interfaces:
            # turn the interface spec into an instance
            lgr.log(5, "Importing module %s " % _intfspec[0])
            _mod = import_module(_intfspec[0], package='niceman')
            _intf = getattr(_mod, _intfspec[1])
            cmd_name = get_cmdline_command_name(_intfspec)
            # deal with optional parser args
            if hasattr(_intf, 'parser_args'):
                parser_args = _intf.parser_args
            else:
                parser_args = dict(formatter_class=formatter_class)
            # use class description, if no explicit description is available
                parser_args['description'] = alter_interface_docs_for_cmdline(
                    _intf.__doc__)
            # create subparser, use module suffix as cmd name
            subparser = subparsers.add_parser(cmd_name, add_help=False, **parser_args)
            # all subparser can report the version
            helpers.parser_add_common_opt(
                subparser, 'version',
                version='niceman %s %s\n\n%s' % (cmd_name, niceman.__version__,
                                                 _license_info()))
            # our own custom help for all commands
            helpers.parser_add_common_opt(subparser, 'help')
            helpers.parser_add_common_opt(subparser, 'log_level')
            # let module configure the parser
            _intf.setup_parser(subparser)
            # logger for command

            # configure 'run' function for this command
            plumbing_args = dict(
                func=_intf.call_from_parser,
                logger=logging.getLogger(_intf.__module__),
                subparser=subparser)
            if hasattr(_intf, 'result_renderer_cmdline'):
                plumbing_args['result_renderer'] = _intf.result_renderer_cmdline
            subparser.set_defaults(**plumbing_args)
            # store short description for later
            sdescr = getattr(_intf, 'short_description',
                             parser_args['description'].split('\n')[0])
            cmd_short_descriptions.append((cmd_name, sdescr))
            parts[cmd_name] = subparser
        grp_short_descriptions.append(cmd_short_descriptions)

    # create command summary
    cmd_summary = []
    for i, grp in enumerate(interface_groups):
        grp_descr = grp[1]
        grp_cmds = grp_short_descriptions[i]

        cmd_summary.append('\n*%s*\n' % (grp_descr,))
        for cd in grp_cmds:
            cmd_summary.append('  - %s:  %s'
                               % (cd[0],
                                  textwrap.fill(
                                      cd[1].rstrip(' .'),
                                      75,
                                      #initial_indent=' ' * 4,
                                      subsequent_indent=' ' * 8)))
    # we need one last formal section to not have the trailed be
    # confused with the last command group
    cmd_summary.append('\n*General information*\n')
    parser.description = '%s\n%s\n\n%s' \
        % (parser.description,
           '\n'.join(cmd_summary),
           textwrap.fill(dedent_docstring("""\
    Detailed usage information for individual commands is
    available via command-specific --help, i.e.:
    niceman <command> --help"""),
                         75, initial_indent='', subsequent_indent=''))
    parts['niceman'] = parser
    lgr.log(5, "Finished setup_parser")
    if return_subparsers:
        return parts
    else:
        return parser


# yoh: arn't used
# def generate_api_call(cmdlineargs=None):
#     parser = setup_parser()
#     # parse cmd args
#     cmdlineargs = parser.parse_args(cmdlineargs)
#     # convert cmdline args into API call spec
#     functor, args, kwargs = cmdlineargs.func(cmdlineargs)
#     return cmdlineargs, functor, args, kwargs


def main(args=None):
    lgr.log(5, "Starting main(%r)", args)
    # PYTHON_ARGCOMPLETE_OK
    parser = setup_parser()
    try:
        import argcomplete
        argcomplete.autocomplete(parser)
    except ImportError:
        pass

    # parse cmd args
    cmdlineargs = parser.parse_args(args)
    if not cmdlineargs.change_path is None:
        for path in cmdlineargs.change_path:
            chpwd(path)

    if cmdlineargs.config is not None:
        # In this case, we're unnecessarily instantiating ConfigManager twice,
        # at import and now, but it might not be worth the effort to
        # restructure things to delay the import.
        from niceman.config import ConfigManager
        niceman.cfg = ConfigManager(cmdlineargs.config, load_default=False)

    if not hasattr(cmdlineargs, 'func'):
        lgr.info("No command given, returning")
        return

    ret = None
    if cmdlineargs.common_debug or cmdlineargs.common_idebug:
        # so we could see/stop clearly at the point of failure
        setup_exceptionhook(ipython=cmdlineargs.common_idebug)
        ret = cmdlineargs.func(cmdlineargs)
    else:
        # otherwise - guard and only log the summary. Postmortem is not
        # as convenient if being caught in this ultimate except
        try:
            ret = cmdlineargs.func(cmdlineargs)
        except InsufficientArgumentsError as exc:
            # if the func reports inappropriate usage, give help output
            lgr.error('%s (%s)' % (exc_str(exc), exc.__class__.__name__))
            cmdlineargs.subparser.print_usage()
            sys.exit(1)
        except MissingConfigFileError as exc:
            # TODO: ConfigManager is not finding files in the default locations.
            lgr.error('%s (%s)' % (exc_str(exc), exc.__class__.__name__))
            sys.exit(1)
        except Exception as exc:
            # print('%s (%s)' % (exc_str(exc), exc.__class__.__name__))
            lgr.error('%s (%s)' % (exc_str(exc), exc.__class__.__name__))
            sys.exit(1)
    if hasattr(cmdlineargs, 'result_renderer'):
        return cmdlineargs.result_renderer(ret)

lgr.log(5, "Done importing cmdline.main")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper utils to deal with various file formats
"""

from __future__ import absolute_import

import yaml

from niceman.utils import safe_write


def write_config_key(stream, envconfig, key, intro_comment=""):
    """Writes the YAML representation of a single key

    This writes a single key of a dict to an output stream and then removes
    the key from the dict.

    Parameters
    ----------
    stream
        Output Stream
    envconfig : dict
    key
        Key from the dictionary
    intro_comment : str, optional
        Introduction comment for the key

    """
    if key in envconfig:
        mini_config = dict()
        mini_config[key] = envconfig.pop(key)
        if intro_comment:
            safe_write(stream, "\n# %s\n\n" % intro_comment)
        write_config(stream, mini_config)


def write_config(stream, rec):
    """TODO"""
    return safe_write(
        stream,
        yaml.safe_dump(
            rec, encoding="utf-8", allow_unicode=True,
            default_flow_style=False
        )
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import datetime
import docker
import logging
import os
from fabric import Connection
import pytest
import tempfile
import uuid

from ..session import get_updated_env, Session
from ...support.exceptions import CommandError
from ..docker_container import DockerSession, PTYDockerSession
from ...utils import chpwd, swallow_logs
from ..shell import ShellSession
from ..singularity import Singularity, SingularitySession, \
    PTYSingularitySession
from ..ssh import SSHSession, PTYSSHSession
from ...tests.utils import skip_ssh
from ...tests.fixtures import get_docker_fixture
from ...consts import TEST_SSH_DOCKER_DIGEST


# Note: due to skip_ssh right here, it would skip the entire module with
# all the tests here if no ssh testing is requested
testing_container = skip_ssh(get_docker_fixture)(
    TEST_SSH_DOCKER_DIGEST,
    name='testing-container',
    portmaps={
        49000: 22
    },
    custom_params={
        'host': 'localhost',
        'user': 'root',
        'port': 49000,
    },
    scope='module'
)


@pytest.mark.skip(reason="TODO")
def test_check_envvars_handling():
    # TODO: test that all the handling of variables works with set_envvar
    # get_envvar etc
    pass


# TODO: make it into a fixture I guess if needed, or just import and call within
# specific backend tests
def check_session_passing_envvars(session):
    # TODO: do not set/pass any env variables, test that PATH is set within remote
    default_envvars = session.query_envvars()
    assert default_envvars['PATH']

    assert 'EXPORTED' not in session.query_envvars()
    session.set_envvar('EXPORTED_NOT_PERMANENT', 'VALUE')
    assert session.query_envvars()['EXPORTED_NOT_PERMANENT'] == 'VALUE'

    session.set_envvar('EXPORTED_PERMANENT', 'VALUE2')
    assert session.query_envvars()['EXPORTED_NOT_PERMANENT'] == 'VALUE'
    assert session.query_envvars()['EXPORTED_PERMANENT'] == 'VALUE2'

    # TODO: we should add functionality to record the state of the env
    # upon finishing create (or install? login?) and here could test
    # smth like
    #  session = session.restart()
    #  envvars = assert session.query_envvars()
    #  assert 'EXPORTED_NOT_PERMANENT' not in envvars
    #  assert envvars['EXPORTED_NOT_PERMANENT'] == 'VALUE2'


def test_get_updated_env():
    assert get_updated_env({'a': 1}, {'a': 2}) == {'a': 2}
    assert get_updated_env({'a': None}, {'a': 2}) == {'a': 2}
    assert get_updated_env({'a': 1}, {'a': None}) == {}
    assert get_updated_env({'a': 1, 'b': 2}, {'a': None}) == {'b': 2}
    assert get_updated_env({'a': 1, 'b': 2}, {'a': None, 'b': 3}) == {'b': 3}


def test_get_local_session():
    # get_local_session(env={'LC_ALL': 'C'}, pty=False, shared=None)
    return


def test_session_class():

    with Session() as session:

        # Check class constructor
        assert type(session) == Session

        # Check __call__ is passing through to _execute_command()
        with pytest.raises(NotImplementedError):
            stdout, stderr = session(['ls'], env={'ENV_VAR': 'ENV_VALUE'})

        session._env = { 'VAR': 'VAR_VALUE' }
        session._env_permanent = { 'PERM_VAR': 'PERM_VAR_VALUE' }

        # Check we can read _envs properly
        envvars = session.get_envvars()
        assert envvars['VAR'] == 'VAR_VALUE'
        assert 'PERM_VAR' not in envvars

        # Check we can read permanent envs
        envvars = session.get_envvars(permanent=True)
        assert envvars['PERM_VAR'] == 'PERM_VAR_VALUE'
        assert 'VAR' not in envvars

        # Check we can add an env through the setter
        session.set_envvar('NEW_VAR', value='NEW_VAR_VALUE')
        envvars = session.get_envvars()
        assert envvars['NEW_VAR'] == 'NEW_VAR_VALUE'
        assert envvars['VAR'] == 'VAR_VALUE'

        # Check we can set an env var by passing a dict
        session.set_envvar({'DICT_VAR': 'DICT_VAR_VALUE'})
        envvars = session.get_envvars()
        assert envvars['DICT_VAR'] == 'DICT_VAR_VALUE'
        assert envvars['VAR'] == 'VAR_VALUE'

        # Check we can delete an existing env var
        session.set_envvar('DICT_VAR', None)
        envvars = session.get_envvars()
        assert 'DICT_VAR' not in envvars
        assert envvars['VAR'] == 'VAR_VALUE'

        # Check formatting of env values
        session.set_envvar('VAR', 'FORMATTED {}', format=True)
        envvars = session.get_envvars()
        assert envvars['VAR'] == 'FORMATTED VAR_VALUE' 
        assert envvars['NEW_VAR'] == 'NEW_VAR_VALUE'

        # At this time, setting permanent env vars is not supported
        with pytest.raises(NotImplementedError):
            session.set_envvar('NEW_VAR', value='NEW_VAR_VALUE',
                permanent=True)

        # Check we raise an exception if user tries to set an env value while
        # passing a dict
        with pytest.raises(AssertionError):
            session.set_envvar({'WILL': 'FAIL'}, value='!')

        # Check query_envvars() method not implemented
        with pytest.raises(NotImplementedError):
            session.query_envvars()

        # Check source_script() method not implemented
        with pytest.raises(NotImplementedError):
            session.source_script(['ls'])

        # Check unauthorized commands raise CommandError exception
        with pytest.raises(CommandError):
            session.niceman_exec('sudo', ['rm', '-r', '/'])

        # Check mangled arg raises exception
        with pytest.raises(CommandError):
            session.niceman_exec('mkdir', ['bad=passed=argument'])

        # Check exec command to valid method passes through to not implemented
        # exception
        with pytest.raises(NotImplementedError):
            session.niceman_exec('mkdir', ['/my/new/dir', 'parents=True'])

        # Check abstract methods raise NotImplementedError
        with pytest.raises(NotImplementedError):
            session.exists('/path')
        with pytest.raises(NotImplementedError):
            session.put('src_path', 'dest_path')
        with pytest.raises(NotImplementedError):
            session.get('src_path', 'dest_path')
        with pytest.raises(NotImplementedError):
            session.get_mtime('path')
        with pytest.raises(NotImplementedError):
            session.read('path')
        with pytest.raises(NotImplementedError):
            session.mkdir('path')
        with pytest.raises(NotImplementedError):
            session.mktmpdir()
        with pytest.raises(NotImplementedError):
            session.isdir('path')
        with pytest.raises(NotImplementedError):
            session.chmod('path', 'mode')
        with pytest.raises(NotImplementedError):
            session.chown('path', 100)


# Run the abstract method tests for each class the inherits the Session class.
@pytest.fixture(params=[
    DockerSession,
    PTYDockerSession,
    ShellSession,
    SingularitySession,
    PTYSingularitySession,
    SSHSession,
    PTYSSHSession
])
def resource_session(request):
    """Pytest fixture that provides instantiated session objects for each
    of the classes that inherit the Session or POSIXSession classes.

    The fixture will run the test method once for each session object provided.

    Parameters
    ----------
    request : object
        Pytest request object that contains the class to test against

    Yields
    -------
    session object
        Instantiated object based on a class that extends the Session or
        POSIXSession class
    """
    # Initialize docker connection to testing container.
    if request.param in [DockerSession, PTYDockerSession]:
        client = docker.Client()
        container = [
            c for c in client.containers()
            if '/testing-container' in c['Names']
        ][0]
        yield request.param(client, container)
    elif request.param in [SSHSession, PTYSSHSession]:
        # Initialize SSH connection to testing Docker container.
        connection = Connection(
            'localhost',
            user='root',
            port=49000,
            connect_kwargs={
                'password': 'root'
            }
        )
        connection.open()
        yield request.param(connection)
    elif request.param in [SingularitySession, PTYSingularitySession]:
        # Initialize Singularity test container.
        name = str(uuid.uuid4().hex)[:11]
        resource = Singularity(name=name, image='docker://python:2.7')
        resource.connect()
        resource.create()
        yield request.param(name)
        resource.delete()
    else:
        yield request.param()



def test_session_abstract_methods(testing_container, resource_session,
        resource_test_dir):

    session = resource_session

    # Check the validity of the env vars
    envs = session.query_envvars()
    assert 'HOME' in envs
    assert 'PATH' in envs

    # Check sourcing new env variables
    # new_envs = session.source_script(['export', 'SCRIPT_VAR=SCRIPT_VALUE'])
    # assert 'SCRIPT_VAR' in new_envs

    # Check _execute_command by checking file system
    out, err = session._execute_command(['cat', '/etc/hosts'])
    assert '127.0.0.1' in out
    assert 'localhost' in out
    assert err == ''

    # Check _execute_command failure
    with pytest.raises(CommandError):
        session._execute_command(['cat', '/no/such/file'])

    # Check _execute_command with env set
    # TODO: Implement env parameter for _execute_command()
    if session.__class__.__name__ != 'ShellSession':
        with pytest.raises(NotImplementedError):
            out, err = session._execute_command(['cat', '/etc/hosts'],
                env={'NEW_VAR': 'NEW_VAR_VALUE'})

    # Check _execute_command with cwd set
    # TODO: Implement cwd parameter for _execute_command()
    if 'DockerSession' in session.__class__.__name__:
        with swallow_logs(new_level=logging.WARN) as log:
            session._execute_command(['cat', '/etc/hosts'], cwd='/tmp')
            assert "cwd is not handled in docker yet" in log.out
    elif session.__class__.__name__ != 'ShellSession':
        with pytest.raises(NotImplementedError):
            out, err = session._execute_command(['cat', '/etc/hosts'],
                cwd='/tmp')

    # Check exists() method
    result = session.exists('/etc')
    assert result
    result = session.exists('/etc/hosts')
    assert result
    result = session.exists('/no/such/file')
    assert not result
    # exists() doesn't get confused by an empty string.
    assert not session.exists('')

    # Check isdir() method
    result = session.isdir('/etc')
    assert result
    result = session.isdir('/etc/hosts')  # A file, not a dir
    assert not result
    result = session.isdir('/no/such/dir')
    assert not result

    # Create a temporary test file
    temp_file = tempfile.NamedTemporaryFile(dir=resource_test_dir)
    with temp_file as f:
        f.write('NICEMAN test content\nline 2\nline 3'.encode('utf8'))
        f.flush()
        local_path = temp_file.name
        remote_path = '{}/niceman upload/{}'.format(resource_test_dir,
            uuid.uuid4().hex)

        # Check put() method
        # session.put(local_path, remote_path, uid=3, gid=3) # UID for sys, GID for sys
        # TODO: Sort out permissions issues with chown for SSH when no sudo
        session.put(local_path, remote_path)
        result = session.exists(remote_path)
        assert result
        # TODO: Check uid and gid of remote file

        # We can use a relative name for the target
        basename_put_dir = os.path.join(resource_test_dir, "basename-put")
        if not os.path.exists(basename_put_dir):
            os.mkdir(basename_put_dir)
        # Change directory to avoid polluting test directory for local shell.
        with chpwd(basename_put_dir):
            try:
                session.put(local_path, os.path.basename(remote_path))
            except ValueError:
                # Docker and Singularity don't accept non-absolute paths.
                assert isinstance(session, (DockerSession, SingularitySession))

    # Check get_mtime() method by checking new file has today's date
    result = int(session.get_mtime(remote_path).split('.')[0])
    assert datetime.datetime.fromtimestamp(result).month == \
        datetime.date.today().month
    assert datetime.datetime.fromtimestamp(result).day == \
        datetime.date.today().day

    # Check read() method
    output = session.read(remote_path).split('\n')
    assert output[0] == 'NICEMAN test content'
    assert output[1] == 'line 2'

    # Check get() method
    local_path = '{}/download/{}'.format(resource_test_dir,
        uuid.uuid4().hex)
    session.get(remote_path, local_path)
    # TODO: In some cases, updating uid and gid does not work if not root
    assert os.path.isfile(local_path)
    with open(local_path, 'r') as f:
        content = f.read().split('\n')
        assert content[0] == 'NICEMAN test content'
    os.remove(local_path)
    os.rmdir(os.path.dirname(local_path))

    with chpwd(resource_test_dir):
        # We can get() without a leading directory.
        session.get(remote_path, "just base")
        assert os.path.exists("just base")
        remote_basename = os.path.basename(remote_path)
        # We can get() without specifying a target.
        session.get(remote_path)
        assert os.path.exists(remote_basename)
        # Or by specifying just the directory.
        session.get(remote_path, "subdir" + os.path.sep)
        assert os.path.exists(os.path.join("subdir", remote_basename))

    # Check mkdir() method
    test_dir = '{}/{}'.format(resource_test_dir, uuid.uuid4().hex)
    session.mkdir(test_dir)
    result = session.isdir(test_dir)
    assert result

    # Check making parent dirs without setting flag
    test_dir = '{}/tmp/i fail/{}'.format(resource_test_dir, uuid.uuid4().hex)
    with pytest.raises(CommandError):
        session.mkdir(test_dir, parents=False)
    result = session.isdir(test_dir)
    assert not result
    # Check making parent dirs when parents flag set
    test_dir = '{}/i succeed/{}'.format(resource_test_dir, uuid.uuid4().hex)
    session.mkdir(test_dir, parents=True)
    result = session.isdir(test_dir)
    assert result

    # Check mktmpdir() method
    test_dir = session.mktmpdir()
    result = session.isdir(test_dir)
    assert result, "The path %s is not a directory" % test_dir

    # All sessions will take the command in string form...
    output_string = "{}/stringtest {}".format(
        resource_test_dir, session.__class__.__name__)
    assert not session.exists(output_string)
    session.execute_command("touch '{}'".format(output_string))
    assert session.exists(output_string)
    # and the list form.
    output_list = "{}/listtest {}".format(
        resource_test_dir, session.__class__.__name__)
    assert not session.exists(output_list)
    session.execute_command(["touch", output_list])
    assert session.exists(output_list)

    # TODO: How to test chmod and chown? Need to be able to read remote file attributes
    # session.chmod(self, path, mode, recursive=False):
    # session.chown(self, path, uid=-1, gid=-1, recursive=False, remote=True):



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Orchestrators helping with management of target environments (remote or local)"""

import os.path as op
import abc
import attr
import collections
import yaml

from importlib import import_module
from six import add_metaclass
from six import viewvalues

from niceman.utils import attrib
from niceman.resource.session import get_local_session

import logging
lgr = logging.getLogger('niceman.distributions')


Factory = attr.Factory


def TypedList(type_):
    """A helper to generate an attribute which would be with list factory 
    but also defining a type in its metadata
    """
    return attrib(default=Factory(list), metadata={'type': type_})


#
# Models
#

class SpecObject(object):
    @staticmethod
    def yaml_representer(dumper, data):

        ordered_items = filter(
            lambda i: bool(i[1]),  # so only non empty/None
            attr.asdict(
                data, dict_factory=collections.OrderedDict).items())
        return dumper.represent_mapping('tag:yaml.org,2002:map', ordered_items)


def _register_with_representer(cls):
    # TODO: check if we could/should just inherit from  yaml.YAMLObject
    # or could may be craft our own metaclass
    yaml.SafeDumper.add_representer(cls, SpecObject.yaml_representer)


@attr.s
class Package(SpecObject):
    # files used/associated with the package
    # Unfortunately cannot be the one with default value in the super-class :-/
    # https://github.com/python-attrs/attrs/issues/38
    # So for now will be defined specifically per each subclass
    # files = attr.ib(default=attr.Factory(list))

    @property
    def _cmp_id(self):
        return tuple(getattr(self, a) for a in self._cmp_fields)


@add_metaclass(abc.ABCMeta)
@attr.s
class Distribution(SpecObject):
    """Base class for distributions"""

    # Actually might want/need to go away since somewhat duplicates the class
    # name and looks awkward
    name = attrib(default=attr.NOTHING)

    @staticmethod
    def factory(distribution_type, provenance=None):
        """
        Factory method for creating the appropriate Orchestrator sub-class
        based on format type.

        Parameters
        ----------
        distribution_type : string
            Type of distribution subclass to create. Current options are:
            'conda', 'debian', 'neurodebian', 'pypi', 'redhat'
        provenance : dict
            Keyword args to be passed to initialize class instance 

        Returns
        -------
        distribution : object
            Distribution class or its instance (when provenance is not None)
        """
        # Handle distributions that don't follow the assumed naming structure.
        special_dists = {"svn": "SVNDistribution"}
        special_modules = {"git": "vcs", "svn": "vcs"}

        dlower = distribution_type.lower()
        class_name = special_dists.get(dlower,
                                       dlower.capitalize() + 'Distribution')
        module = import_module('niceman.distributions.' +
                               special_modules.get(dlower, dlower))
        class_ = getattr(module, class_name)
        return class_ if provenance is None else class_(**provenance)

    @abc.abstractmethod
    def initiate(self, session):
        """
        Perform any initialization commands needed in the environment.

        Parameters
        ----------
        session : object
            The Session to work in
        """
        return

    @abc.abstractmethod
    def install_packages(self, session=None):
        """
        Install the packages associated to this distribution by the provenance
        into the environment.

        Parameters
        ----------
        session : object
            Session to work in
        """
        return

# So this one is no longer "distributions/" module specific
# TODO: move up! and strip Spec suffix
@attr.s
class EnvironmentSpec(SpecObject):
    base = attrib()  # ???  to define specifics of the system, possibly a docker base
    distributions = TypedList(Distribution)  # list of distributions
    files = attrib(default=Factory(list))  # list of other files
    # runs?  whenever we get to provisioning executions
    #        those would also be useful for tracing for presence of distributions
    #        e.g. depending on what is in the PATH


    def get_distribution(self, dtype):
        """get_distribution(dtype) -> distribution

        Returns the distribution of the specified type in the given 
        environment.  Returns None if there are no matching distributions.  
        Raises ValueError if there is more than one matching distribution.
        """
        dist = None
        for d in self.distributions:
            if isinstance(d, dtype):
                if dist:
                    raise ValueError('multiple %s found' % str(dtype))
                dist = d
        return dist

_register_with_representer(EnvironmentSpec)


# Note: The following was derived from ReproZip's PkgManager class
# (Revised BSD License)

@add_metaclass(abc.ABCMeta)
class DistributionTracer(object):
    """Base class for package trackers.

    ATM :term:`Package` describes all of possible "archives" which deliver
    some piece of software or data -- packages by distributions (Debian, conda,
    pip, ...), VCS repositories or even container images -- something which has
    to be installed to fulfill environment spec
    """

    # Default to being able to handle directories
    HANDLES_DIRS = True

    def __init__(self, session=None):
        # will be (re)used to run external commands, and let's hardcode LC_ALL
        # codepage just in case since we might want to comprehend error
        # messages
        self._session = session or get_local_session()
        # to ease _init within derived classes which should not be parametrized
        # more anyways
        self._init()

    def _init(self):
        pass

    @abc.abstractmethod
    def identify_distributions(self, files):
        return

    # This one assumes that distribution works with "packages"
    # TODO: we might want to create a more specialized sub-class for that purpose
    # and move those methods below into that subclass
    # TODO: moreover this one assumes only detection by the files and
    #       provides a generic implementation which is based on stages:
    #       1. for each file identifying package fields uniquely describing the
    #          package (method `_get_packagefields_for_files`)
    #       2. grouping fields based on the packagefields
    #       3. creating an actual `Package` using those fields for each group
    #          of files.
    #  In principle could be RFed to be more scalable, where there is a "Package
    #  manager", which provides similar "assign file to a package" functionality
    #  and identifying packages as already known to the manager.
    def identify_packages_from_files(self, files, root_key=None):
        """Identifies "packages" for a given collection of files

        From an iterative collection of files, we identify the packages
        that contain the files and any files that are not related.

        Parameters
        ----------
        files : iterable
            Container (e.g. list or set) of file paths
        root_key : string, optional
            When adding a matched file to the returned package dict, represent
            that path as relative to the value of this package field rather a
            full path.

        Return
        ------
        (found_packages, unknown_files)
            - found_packages is a list of dicts that holds information about
              the found packages. Package dicts need at least "name" and
              "files" (that contains an array of related files)
            - unknown_files is a list of files that were not found in
              a package
        """
        unknown_files = set()
        found_packages = {}
        nb_pkg_files = 0

        # TODO: probably that _get_packagefields should create packagespecs
        # internally and just return them.  But we should make them hashable
        file_to_package_dict = self._get_packagefields_for_files(files)
        for f in files:
            # Stores the file
            if f not in file_to_package_dict:
                unknown_files.add(f)
            else:
                pkgfields = file_to_package_dict[f]
                if pkgfields is None:
                    unknown_files.add(f)
                else:
                    if root_key:
                        f_pkg = op.relpath(f, pkgfields[root_key])
                    else:
                        f_pkg = f
                    pkgfields_hashable = tuple(x for x in pkgfields.items())
                    if pkgfields_hashable in found_packages:
                        found_packages[pkgfields_hashable].files.append(f_pkg)
                        nb_pkg_files += 1
                    else:
                        pkg = self._create_package(**pkgfields)
                        if pkg:
                            found_packages[pkgfields_hashable] = pkg
                            # we store only non-directories within 'files'
                            if not self._session.isdir(f):
                                pkg.files.append(f_pkg)
                            nb_pkg_files += 1
                        else:
                            unknown_files.add(f)

        lgr.debug(
            "%s: %d packages with %d files, and %d other files",
            self.__class__.__name__,
            len(found_packages),
            nb_pkg_files,
            len(unknown_files))

        return list(viewvalues(found_packages)), unknown_files

    @abc.abstractmethod
    def _get_packagefields_for_files(self, files):
        """Given a list of files, should return a dict mapping files to a
        dictionary of fields which would be later passed into _create_package
        to actually create packages while grouping into packages
        (having identical returned packagefield values)
        """
        return

    @abc.abstractmethod
    def _create_package(self, **package_fields):
        """Creates implementation-specific Package object using fields
        provided by _get_packagefields_for_files
        """
        return



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper utility to delete an environment
"""

__docformat__ = 'restructuredtext'

import re

from .base import Interface
import niceman.interface.base # Needed for test patching
from .common_opts import resref_arg
from .common_opts import resref_type_opt
from ..support.param import Parameter
from ..support.constraints import EnsureStr
from ..resource import get_manager

from logging import getLogger
lgr = getLogger('niceman.api.start')


class Start(Interface):
    """Start a computation environment

    Examples
    --------

      $ niceman start my-resource

    """

    _params_ = dict(
        resref=resref_arg,
        resref_type=resref_type_opt,
    )

    @staticmethod
    def __call__(resref, resref_type="auto"):
        manager = get_manager()
        resource = manager.get_resource(resref, resref_type)
        resource.connect()
        manager.start(resource)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Miscellaneous utilities to assist with testing"""

import inspect
import os
import re
import tempfile
import platform
import multiprocessing
import logging
from six import PY2, text_type
from mock import patch
import pytest

from niceman.support.external_versions import external_versions
from six.moves.SimpleHTTPServer import SimpleHTTPRequestHandler
from six.moves.BaseHTTPServer import HTTPServer
from six import reraise

from functools import wraps
from os.path import exists, realpath, join as opj

from ..cmd import Runner
from ..utils import *
from ..dochelpers import borrowkwargs
from ..resource.docker_container import DockerContainer

# temp paths used by clones
_TEMP_PATHS_CLONES = set()


# pytest variants for nose.tools commands.  These exist to avoid unnecessary
# churn in tests that already use these names.  New code should use plain
# asserts to take advantage of pytest's assertion introspection.


def assert_equal(a, b, msg=None):
    assert a == b, msg or "{!r} != {!r}".format(a, b)


def assert_not_equal(a, b, msg=None):
    assert a != b, msg or "{!r} == {!r}".format(a, b)


def assert_greater(a, b, msg=None):
    assert a > b, msg or "{!r} > {!r}".format(a, b)


def assert_greater_equal(a, b, msg=None):
    assert a >= b, msg or "{!r} >= {!r}".format(a, b)


def assert_true(x, msg=None):
    assert x, msg or "{!r} is not true".format(x)


def assert_false(x, msg=None):
    assert not x, msg or "{!r} is not false".format(x)


def assert_in(x, collection, msg=None):
    assert x in collection, \
        msg or "{!r} not found in {!r}".format(x, collection)


def assert_not_in(x, collection, msg=None):
    assert x not in collection, \
        msg or "{!r} unexpectedly found in {!r}".format(x, collection)


def assert_is(a, b, msg=None):
    assert a is b, msg or "{!r} is not {!r}".format(a, b)


def assert_is_instance(a, b, msg=None):
    assert isinstance(a, b), \
        msg or "{!r} is not an instance of {!r}".format(a, b)


# additional shortcuts
assert_raises = pytest.raises
eq_ = assert_equal
neq_ = assert_not_equal
ok_ = assert_true
nok_ = assert_false
in_ = assert_in


# def create_tree_archive(path, name, load, overwrite=False, archives_leading_dir=True):
#     """Given an archive `name`, create under `path` with specified `load` tree
#     """
#     from ..support.archives import compress_files
#     dirname = file_basename(name)
#     full_dirname = opj(path, dirname)
#     os.makedirs(full_dirname)
#     create_tree(full_dirname, load, archives_leading_dir=archives_leading_dir)
#     # create archive
#     if archives_leading_dir:
#         compress_files([dirname], name, path=path, overwrite=overwrite)
#     else:
#         compress_files(list(map(basename, glob.glob(opj(full_dirname, '*')))),
#                        opj(pardir, name),
#                        path=opj(path, dirname),
#                        overwrite=overwrite)
#     # remove original tree
#     shutil.rmtree(full_dirname)


def create_tree(path, tree, archives_leading_dir=True):
    """Given a list of tuples (name, load) create such a tree

    if load is a tuple itself -- that would create either a subtree or an archive
    with that content and place it into the tree if name ends with .tar.gz
    """
    lgr.log(5, "Creating a tree under %s", path)
    if not exists(path):
        os.makedirs(path)

    if isinstance(tree, dict):
        tree = tree.items()

    for name, load in tree:
        full_name = opj(path, name)
        if isinstance(load, (tuple, list, dict)):
            # if name.endswith('.tar.gz') or name.endswith('.tar'):
            #     create_tree_archive(path, name, load, archives_leading_dir=archives_leading_dir)
            # else:
                create_tree(full_name, load, archives_leading_dir=archives_leading_dir)
        else:
            #encoding = sys.getfilesystemencoding()
            #if isinstance(full_name, text_type):
            #    import pydb; pydb.debugger()
            with open(full_name, 'w') as f:
                if PY2 and isinstance(load, text_type):
                    load = load.encode('utf-8')
                f.write(load)

#
# Addition "checkers"
#

#
# Helpers to test symlinks
#

def ok_symlink(path):
    """Checks whether path is either a working or broken symlink"""
    link_path = os.path.islink(path)
    if not link_path:
        raise AssertionError("Path {} seems not to be a symlink".format(path))


def ok_good_symlink(path):
    ok_symlink(path)
    rpath = realpath(path)
    ok_(exists(rpath),
        msg="Path {} seems to be missing.  Symlink {} is broken".format(
                rpath, path))


def ok_broken_symlink(path):
    ok_symlink(path)
    rpath = realpath(path)
    assert_false(exists(rpath),
            msg="Path {} seems to be present.  Symlink {} is not broken".format(
                    rpath, path))


def ok_startswith(s, prefix):
    ok_(s.startswith(prefix),
        msg="String %r doesn't start with %r" % (s, prefix))


def ok_endswith(s, suffix):
    ok_(s.endswith(suffix),
        msg="String %r doesn't end with %r" % (s, suffix))


def nok_startswith(s, prefix):
    assert_false(s.startswith(prefix),
        msg="String %r starts with %r" % (s, prefix))


def ok_generator(gen):
    assert_true(inspect.isgenerator(gen), msg="%s is not a generator" % gen)


def ok_file_has_content(path, content):
    """Verify that file exists and has expected content"""
    assert(exists(path))
    with open(path, 'r') as f:
        assert_equal(f.read(), content)


def assert_in_in(substr, lst):
    """Verify that a substring is in an element of a list"""
    for s in lst:
        if substr in s:
            return
    assert False, '"%s" is not in "%s"' % (substr, str(lst))

#
# Decorators
#

@optional_args
def with_tree(t, tree=None, archives_leading_dir=True, delete=True, **tkwargs):

    @wraps(t)
    def newfunc(*arg, **kw):
        tkwargs_ = get_tempfile_kwargs(tkwargs, prefix="tree", wrapped=t)
        d = tempfile.mkdtemp(**tkwargs_)
        create_tree(d, tree, archives_leading_dir=archives_leading_dir)
        try:
            return t(*(arg + (d,)), **kw)
        finally:
            if delete:
                rmtemp(d)
    return newfunc


lgr = logging.getLogger('niceman.tests')


class SilentHTTPHandler(SimpleHTTPRequestHandler):
    """A little adapter to silence the handler
    """
    def __init__(self, *args, **kwargs):
        self._silent = lgr.getEffectiveLevel() > logging.DEBUG
        SimpleHTTPRequestHandler.__init__(self, *args, **kwargs)

    def log_message(self, format, *args):
        if self._silent:
            return
        lgr.debug("HTTP: " + format % args)


def _multiproc_serve_path_via_http(hostname, path_to_serve_from, queue): # pragma: no cover
    chpwd(path_to_serve_from)
    httpd = HTTPServer((hostname, 0), SilentHTTPHandler)
    queue.put(httpd.server_port)
    httpd.serve_forever()


@optional_args
def serve_path_via_http(tfunc, *targs):
    """Decorator which serves content of a directory via http url
    """

    @wraps(tfunc)
    def newfunc(*args, **kwargs):

        if targs:
            # if a path is passed into serve_path_via_http, then it's in targs
            assert len(targs) == 1
            path = targs[0]

        elif len(args) > 1:
            args, path = args[:-1], args[-1]
        else:
            args, path = (), args[0]

        # There is a problem with Haskell on wheezy trying to
        # fetch via IPv6 whenever there is a ::1 localhost entry in
        # /etc/hosts.  Apparently fixing that docker image reliably
        # is not that straightforward, although see
        # http://jasonincode.com/customizing-hosts-file-in-docker/
        # so we just force to use 127.0.0.1 while on wheezy
        #hostname = '127.0.0.1' if on_debian_wheezy else 'localhost'
        hostname = '127.0.0.1'

        queue = multiprocessing.Queue()
        multi_proc = multiprocessing.Process(
            target=_multiproc_serve_path_via_http,
            args=(hostname, path, queue))
        multi_proc.start()
        port = queue.get(timeout=300)
        url = 'http://{}:{}/'.format(hostname, port)
        lgr.debug("HTTP: serving {} under {}".format(path, url))

        try:
            # Such tests don't require real network so if http_proxy settings were
            # provided, we remove them from the env for the duration of this run
            env = os.environ.copy()
            env.pop('http_proxy', None)
            with patch.dict('os.environ', env, clear=True):
                return tfunc(*(args + (path, url)), **kwargs)
        finally:
            lgr.debug("HTTP: stopping server under %s" % path)
            multi_proc.terminate()

    return newfunc


@optional_args
def without_http_proxy(tfunc):
    """Decorator to remove http*_proxy env variables for the duration of the test
    """

    @wraps(tfunc)
    def newfunc(*args, **kwargs):
        # Such tests don't require real network so if http_proxy settings were
        # provided, we remove them from the env for the duration of this run
        env = os.environ.copy()
        env.pop('http_proxy', None)
        env.pop('https_proxy', None)
        with patch.dict('os.environ', env, clear=True):
            return tfunc(*args, **kwargs)

    return newfunc


@borrowkwargs(methodname=make_tempfile)
@optional_args
def with_tempfile(t, **tkwargs):
    """Decorator function to provide a temporary file name and remove it at the end

    Parameters
    ----------

    To change the used directory without providing keyword argument 'dir' set
    NICEMAN_TESTS_TEMPDIR.

    Examples
    --------

    ::

        @with_tempfile
        def test_write(tfile):
            open(tfile, 'w').write('silly test')
    """

    @wraps(t)
    def newfunc(*arg, **kw):
        with make_tempfile(wrapped=t, **tkwargs) as filename:
            return t(*(arg + (filename,)), **kw)

    return newfunc


def skip_if_no_network(func=None):
    """Skip test completely in NONETWORK settings

    If not used as a decorator, and just a function, could be used at the module level
    """

    def check_and_raise():
        if os.environ.get('NICEMAN_TESTS_NONETWORK'):
            pytest.skip("Skipping since no network settings",
                        allow_module_level=True)

    if func:
        @wraps(func)
        def newfunc(*args, **kwargs):
            check_and_raise()
            return func(*args, **kwargs)
        # right away tag the test as a networked test
        tags = getattr(newfunc, 'tags', [])
        newfunc.tags = tags + ['network']
        return newfunc
    else:
        check_and_raise()


def skip_if_on_windows(func):
    """Skip test completely under Windows
    """
    @wraps(func)
    def newfunc(*args, **kwargs):
        if on_windows:
            pytest.skip("Skipping on Windows", allow_module_level=True)
        return func(*args, **kwargs)
    return newfunc


def skip_if_no_apt_cache(func=None):
    """Skip test completely if apt is unavailable

    If not used as a decorator, and just a function, could be used at the module level
    """

    def check_and_raise():
        if not external_versions["cmd:apt-cache"]:
            pytest.skip("Skipping since apt-cache is not available",
                        allow_module_level=True)

    if func:
        @wraps(func)
        def newfunc(*args, **kwargs):
            check_and_raise()
            return func(*args, **kwargs)
        return newfunc
    else:
        check_and_raise()


def skip_if_no_svn():
    runner = Runner()
    try:
        # will raise OSError(errno=2) if the command is not found
        runner.run(['svnadmin', '--help'])
        runner.run(['svn', '--help'])
    except OSError as exc:
        if exc.errno == 2:
            pytest.skip('subversion is not installed',
                        allow_module_level=True)
    return


def skip_ssh(func=None):
    """Skips SSH tests if on windows or if environment variable
    NICEMAN_TESTS_SSH was not set
    """

    def check_and_raise():
        if not os.environ.get('NICEMAN_TESTS_SSH'):
            pytest.skip("Run this test by setting NICEMAN_TESTS_SSH",
                        allow_module_level=True)

    if func:
        @wraps(func)
        def newfunc(*args, **kwargs):
            if on_windows:
                pytest.skip("SSH currently not available on windows.",
                            allow_module_level=True)
            check_and_raise()
            return func(*args, **kwargs)
        return newfunc
    else:
        check_and_raise()


def skip_if_no_docker_container(container_name='testing-container'):
    """Test decorator that will skip a test if the Docker container the test is
    going to connect to is not running in the Docker engine.

    Parameters
    ----------
    container_name : str
        Name of the container that needs to be running for the test to work.

    Returns
    -------
    func
        Decorator function

    Raises
    ------
    SkipTest
    """
    def decorator(func):
        if not DockerContainer.is_container_running(container_name):
            pytest.skip("Docker container '{}' not running, "
                        "skipping test  {}".format(container_name, func.__name__),
                        allow_module_level=True)
        return func
    return decorator


def skip_if_no_docker_engine(func):
    """Test decorator that will skip a test if a Docker engine can't be found.

    Returns
    -------
    func
        Decorator function

    Raises
    ------
    SkipTest
    """
    if not DockerContainer.is_engine_running():
        pytest.skip("Docker not found, skipping test {}".format(func.__name__),
                    allow_module_level=True)
    return func


def skip_if_no_singularity(func):
    """Test decorator that will skip a test if the singularity executable is
    not found.
    
    Returns
    -------
    func
        Decorator function
    """
    # Make sure singularity is installed
    try:
        stdout, _ = Runner().run(['singularity', '--version'])
    except Exception:
        msg = "Singularity not installed, skipping test {}"
        pytest.skip(msg.format(func.__name__), allow_module_level=True)

    if stdout.startswith('2.2') or stdout.startswith('2.3'):
        # Running singularity instances and managing them didn't happen
        # until version 2.4. See: https://singularity.lbl.gov/archive/
        msg = "Singularity version >= 2.4 required, skipping test {}"
        pytest.skip(msg.format(func.__name__), allow_module_level=True)
    return func

@optional_args
def assert_cwd_unchanged(func, ok_to_chdir=False):
    """Decorator to test whether the current working directory remains unchanged

    Parameters
    ----------
    ok_to_chdir: bool, optional
      If True, allow to chdir, so this decorator would not then raise exception
      if chdir'ed but only return to original directory
    """

    @wraps(func)
    def newfunc(*args, **kwargs):
        cwd_before = os.getcwd()
        pwd_before = getpwd()
        exc_info = None
        try:
            func(*args, **kwargs)
        except:
            exc_info = sys.exc_info()
        finally:
            try:
                cwd_after = os.getcwd()
            except OSError as e:
                lgr.warning("Failed to getcwd: %s" % e)
                cwd_after = None

        if cwd_after != cwd_before:
            chpwd(pwd_before)
            if not ok_to_chdir:
                lgr.warning(
                    "%s changed cwd to %s. Mitigating and changing back to %s"
                    % (func, cwd_after, pwd_before))
                # If there was already exception raised, we better re-raise
                # that one since it must be more important, so not masking it
                # here with our assertion
                if exc_info is None:
                    assert_equal(cwd_before, cwd_after,
                                 "CWD changed from %s to %s" % (cwd_before, cwd_after))

        if exc_info is not None:
            reraise(*exc_info)

    return newfunc


@optional_args
def run_under_dir(func, newdir='.'):
    """Decorator to run tests under another directory

    It is somewhat ugly since we can't really chdir
    back to a directory which had a symlink in its path.
    So using this decorator has potential to move entire
    testing run under the dereferenced directory name -- sideeffect.

    The only way would be to instruct testing framework (i.e. nose
    in our case ATM) to run a test by creating a new process with
    a new cwd
    """

    @wraps(func)
    def newfunc(*args, **kwargs):
        pwd_before = getpwd()
        try:
            chpwd(newdir)
            func(*args, **kwargs)
        finally:
            chpwd(pwd_before)


    return newfunc


def assert_re_in(regex, c, flags=0):
    """Assert that container (list, str, etc) contains entry matching the regex
    """
    if not isinstance(c, (list, tuple)):
        c = [c]
    for e in c:
        if re.match(regex, e, flags=flags):
            return
    raise AssertionError("Not a single entry matched %r in %r" % (regex, c))


# List of most obscure filenames which might or not be supported by different
# filesystems across different OSs.  Start with the most obscure
OBSCURE_FILENAMES = (
    " \"';a&b/&cd `| ",  # shouldn't be supported anywhere I guess due to /
    " \"';a&b&cd `| ",
    " \"';abcd `| ",
    " \"';abcd | ",
    " \"';abcd ",
    " ;abcd ",
    " ;abcd",
    " ab cd ",
    " ab cd",
    "a",
    " abc d.dat ",  # they all should at least support spaces and dots
)

@with_tempfile(mkdir=True)
def get_most_obscure_supported_name(tdir):
    """Return the most obscure filename that the filesystem would support under TEMPDIR

    TODO: we might want to use it as a function where we would provide tdir
    """
    for filename in OBSCURE_FILENAMES:
        if on_windows and filename.rstrip() != filename:
            continue
        try:
            with open(opj(tdir, filename), 'w') as f:
                f.write("TEST LOAD")
            return filename  # it will get removed as a part of wiping up the directory
        except:
            lgr.debug("Filename %r is not supported on %s under %s",
                      filename, platform.system(), tdir)
            pass
    raise RuntimeError("Could not create any of the files under %s among %s"
                       % (tdir, OBSCURE_FILENAMES))


@optional_args
def with_testsui(t, responses=None):
    """Switch main UI to be 'tests' UI and possibly provide answers to be used"""

    @wraps(t)
    def newfunc(*args, **kwargs):
        from niceman.ui import ui
        old_backend = ui.backend
        try:
            ui.set_backend('tests')
            if responses:
                ui.add_responses(responses)
            ret = t(*args, **kwargs)
            if responses:
                responses_left = ui.get_responses()
                assert not len(responses_left), "Some responses were left not used: %s" % str(responses_left)
            return ret
        finally:
            ui.set_backend(old_backend)

    return newfunc
with_testsui.__test__ = False


def assert_is_subset_recur(a, b, subset_types=[]):
    """Asserts that 'a' is a subset of 'b' (recursive on dicts and lists)

    Parameters
    ----------
    a : dict or list
        The desired subset collection (items that must be in b)
    b : dict or list
        The superset collection
    subset_types : list
        List of classes (from list, dict) that allow subsets. Otherwise
        we use strict matching.
"""
    # Currently we only allow lists and dicts
    assert {list, dict}.issuperset(subset_types)
    # For dictionaries recursively check children that are in a
    if isinstance(a, dict) and isinstance(b, dict) and dict in subset_types:
        for key in a:
            if key not in b:
                raise AssertionError("Key %s is missing" % key)
            assert_is_subset_recur(a[key], b[key], subset_types)
    # For lists, recurse for every value a to make sure it is in b
    # (note: two items in a may match the same item in b)
    elif isinstance(a, list) and isinstance(b, list) and list in subset_types:
        for a_val in a:
            for b_val in b:
                try:
                    assert_is_subset_recur(a_val, b_val, subset_types)
                    break
                except AssertionError:
                    pass
            else:
                raise AssertionError("Array value %s is missing" % a_val)
    # For anything else check for straight equality
    else:
        if a != b:
            raise AssertionError("Value %s != %s" % (a, b))


def create_pymodule(directory):
    """Create a skeleton Python module in `directory`.

    Parameters
    ----------
    directory : str
        Path to a non-existing directory.
    """
    os.makedirs(directory)
    with open(os.path.join(directory, "setup.py"), "w") as ofh:
        ofh.write("""\
from setuptools import setup

setup(name='nmtest',
      version='0.1.0',
      py_modules=['nmtest'])""")

    with open(os.path.join(directory, "nmtest"), "w") as ofh:
        ofh.write("")


#
# Context Managers
#



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Support for Python's virtualenv."""
from collections import defaultdict
import logging
import os
import os.path as op

import attr

from six import iteritems
from niceman.distributions import Distribution
from niceman.distributions import piputils
from niceman.dochelpers import borrowdoc
from niceman.dochelpers import exc_str
from niceman.utils import attrib, PathRoot, is_subpath
from niceman.utils import execute_command_batch
from niceman.utils import parse_semantic_version
from niceman.resource.session import get_local_session

from .base import DistributionTracer
from .base import Package
from .base import SpecObject
from .base import TypedList

lgr = logging.getLogger("niceman.distributions.venv")


@attr.s
class VenvPackage(Package):
    name = attrib(default=attr.NOTHING)
    version = attrib(default=attr.NOTHING)
    local = attrib(default=False)
    location = attrib()
    editable = attrib(default=False)
    files = attrib(default=attr.Factory(list))


@attr.s
class VenvEnvironment(SpecObject):
    path = attrib()
    python_version = attrib()
    packages = TypedList(VenvPackage)


@attr.s
class VenvDistribution(Distribution):
    """Class to provide virtualenv-based "distributions".
    """
    path = attrib()
    venv_version = attrib()
    environments = TypedList(VenvEnvironment)

    def initiate(self, _):
        return

    @borrowdoc(Distribution)
    def install_packages(self, session=None):
        session = session or get_local_session()
        for env in self.environments:
            # TODO: Deal with system and editable packages.
            to_install = ["{p.name}=={p.version}".format(p=p)
                          for p in env.packages
                          if p.local and not p.editable]
            if not to_install:
                lgr.info("No local, non-editable packages found")
                continue

            # TODO: Right now we just use the python to invoke "virtualenv
            # --python=..." when the directory doesn't exist, but we should
            # eventually use the yet-to-exist "satisfies" functionality to
            # check whether an existing virtual environment has the right
            # python (and maybe other things).
            pyver = "{v.major}.{v.minor}".format(
                v=parse_semantic_version(env.python_version))

            if not session.exists(env.path):
                # The location and version of virtualenv are recorded at the
                # time of tracing, but should we use these values?  For now,
                # use a plain "virtualenv" below on the basis that we just use
                # "apt-get" and "git" elsewhere.
                session.execute_command(["virtualenv",
                                         "--python=python{}".format(pyver),
                                         env.path])
            list(execute_command_batch(session,
                                       [env.path + "/bin/pip", "install"],
                                       to_install))


class VenvTracer(DistributionTracer):
    """Distribution tracer for virtualenv.
    """

    def _init(self):
        self._path_root = PathRoot(self._is_venv_directory)

    def _get_packagefields_for_files(self, files):
        raise NotImplementedError

    def _create_package(self, **package_fields):
        raise NotImplementedError

    def _get_package_details(self, venv_path):
        pip = venv_path + "/bin/pip"
        try:
            packages, file_to_pkg = piputils.get_package_details(
                self._session, pip)
        except Exception as exc:
            lgr.warning("Could not determine pip package details for %s: %s",
                        venv_path, exc_str(exc))
            return {}, {}
        return packages, file_to_pkg

    def _is_venv_directory(self, path):
        try:
            self._session.execute_command(["grep", "-q", "VIRTUAL_ENV",
                                           "{}/bin/activate".format(path)])
        except Exception as exc:
            lgr.debug("Did not detect virtualenv at the path %s: %s",
                      path, exc_str(exc))
            return False
        return True

    def _get_venv_path(self, path):
        return self._path_root(path)

    def identify_distributions(self, files):
        unknown_files = set(files)
        found_package_count = 0

        venv_paths = map(self._get_venv_path, files)
        venv_paths = set(filter(None, venv_paths))

        venvs = []
        for venv_path in venv_paths:
            package_details, file_to_pkg = self._get_package_details(venv_path)
            local_pkgs = set(piputils.get_pip_packages(self._session,
                                                       venv_path + "/bin/pip",
                                                       restriction="local"))
            pkg_to_found_files = defaultdict(list)
            for path in set(unknown_files):  # Clone the set
                # The supplied path may be relative or absolute, but
                # file_to_pkg keys are absolute paths.
                fullpath = os.path.abspath(path)
                if fullpath in file_to_pkg:
                    unknown_files.remove(path)
                    pkg_to_found_files[file_to_pkg[fullpath]].append(
                        os.path.relpath(path, venv_path))

            # Some files, like venvs/dev/lib/python2.7/abc.py could be
            # symlinks populated by virtualenv itself during venv creation
            # since it relies on system wide python environment.  So we need
            # to resolve those into filenames which could be associated with
            # system wide installation of python
            for path in unknown_files.copy():
                if is_subpath(path, venv_path) and op.islink(path):
                    unknown_files.add(op.realpath(path))
                    unknown_files.remove(path)

            packages = []
            for name, details in iteritems(package_details):
                location = details["location"]
                packages.append(
                    VenvPackage(name=details["name"],
                                version=details["version"],
                                local=name in local_pkgs,
                                location=location,
                                editable=details["editable"],
                                files=pkg_to_found_files[name]))
                if location and not is_subpath(location, venv_path):
                    unknown_files.add(location)

            found_package_count += len(packages)

            venvs.append(
                VenvEnvironment(path=venv_path,
                                python_version=self._python_version(venv_path),
                                packages=packages))

        if venvs:
            yield (VenvDistribution(name="venv",
                                    venv_version=self._venv_version(),
                                    path=self._venv_exe_path(),
                                    environments=venvs),
                   unknown_files)

    def _python_version(self, venv_path):
        try:
            out, err = self._session.execute_command(
                [venv_path + "/bin/python", "--version"])
            # Python 2 sends its version to stderr, while Python 3
            # sends it to stdout.
            pyver = out if "Python" in out else err
            return pyver.strip().split()[1]
        except Exception as exc:
            lgr.debug("Could not determine python version: %s",
                      exc_str(exc))
            return

    # A virtualenv directory doesn't contain any information about
    # which virtualenv created it, so we just go with its current
    # version and location.

    def _venv_version(self):
        try:
            out, _ = self._session.execute_command(["virtualenv", "--version"])
        except Exception as exc:
            lgr.debug("Could not determine virtualenv version: %s",
                      exc_str(exc))
            return
        return out.strip()

    def _venv_exe_path(self):
        try:
            out, _ = self._session.execute_command(["which", "virtualenv"])
        except Exception as exc:
            lgr.debug("Could not determine virtualenv path: %s",
                      exc_str(exc))
            return
        return out.strip()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper utility to install packages
"""

__docformat__ = 'restructuredtext'

from .base import Interface
from .common_opts import resref_arg
from .common_opts import resref_type_opt
from ..support.param import Parameter
from ..support.constraints import EnsureStr
from ..formats import Provenance
from ..resource import get_manager

from logging import getLogger
lgr = getLogger('niceman.api.install')


class Install(Interface):
    """Install packages according to the provided specification(s)

    Examples
    --------

      $ niceman install docker recipe_for_failure.yml

    """

    _params_ = dict(
        resref=resref_arg,
        resref_type=resref_type_opt,
        spec=Parameter(
            args=("spec",),
            doc="file with specifications (in supported formats) of"
                " packages used in executed environment",
            metavar='SPEC',
            nargs="+",
            constraints=EnsureStr(),
            # TODO:  here we need to elaborate options for sub-parsers to
            # provide options, like --no-exec, etc  per each spec
            # ACTUALLY this type doesn't work for us since it is --spec SPEC SPEC... TODO
        ),
    )

    @staticmethod
    def __call__(resref, spec, resref_type="auto"):
        # Load, while possible merging/augmenting sequentially
        assert len(spec) == 1, "For now supporting having only a single spec"
        filename = spec[0]
        provenance = Provenance.factory(filename)

        # TODO
        #  - provenance might contain a 'base' which would instruct which
        #    resource to use

        env_resource = get_manager().get_resource(resref, resref_type)
        env_resource.connect()

        #  TODOs:
        #  - resource might be a backend, so we would need to do analysis
        #    to choose a base environment (e.g. docker image) first, given
        #    provenance details
        #  - might need to create a new session if by now we have only an
        #    "image" which cannot be used as a session right away
        #  - eventually we would need to implement an analysis to determine
        #    details of the provenance.distribution(s) before passing them
        #    for initiation/installation.  That would also "kick back" on the
        #    steps above making things "tricky" ;)

        # For now we deal with simple resources providing a session
        # and a complete, exhaustive and non conflicting with the specified
        # resource
        session = env_resource.get_session()
        environment_spec = provenance.get_environment()
        for distribution in environment_spec.distributions:
            # TODO: add option to skip initiation
            distribution.initiate(session)
            distribution.install_packages(session)
        #env_resource.execute_command_buffer()
        # ??? verify that everything was installed according to the specs
        #     so would need pretty much going through the spec and querying
        #     all those packages.  If something differs -- report
        # session.close()
        if environment_spec.files:
            lgr.warning("Got extra files listed %s", environment_spec.files)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Base classes"""

import attr
from .utils import attrib


@attr.s
class BaseSpec(object):
    """Base class for all NICEMAN Specs
    
    Accent is made on providing exhaustive description of the elements
    
    Should be YAMLable
    """
    pass  # nothing in it


@attr.s
class Package(BaseSpec):
    """Base class to represent a "package" as an installable/exchangeable unit
    
    Unit comes with files which were traced to be relevant/important for a given
    computation
    """
    files = attrib(default=[])


@attr.s
class VCS(Package):
    #path = attr.ib()
    pass


@attr.s
class Git(VCS):
    pass








# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
import json
import os
from os.path import islink
from os.path import join, isfile

import logging

import attr

from niceman.distributions.debian import DebTracer
from niceman.distributions.debian import DEBPackage
from niceman.distributions.debian import DebianDistribution

import pytest

import mock

from niceman.utils import swallow_logs
from niceman.tests.utils import skip_if_no_apt_cache


@skip_if_no_apt_cache
def test_dpkg_manager_identify_packages():
    files = ["/sbin/iptables"]
    tracer = DebTracer()
    (packages, unknown_files) = \
        tracer.identify_packages_from_files(files)
    # Make sure that iptables was identified
    assert (not unknown_files), "/sbin/iptables should be identified"
    assert len(packages) == 1
    pkg = packages[0]
    assert pkg.name == 'iptables'
    # Make sure apt_sources are identified, but then we should ask the entire
    # distribution
    distributions = list(tracer.identify_distributions(files))
    assert len(distributions) == 1
    distribution, unknown_files = distributions[0]
    assert distribution.apt_sources
    # Make sure both a non-local origin was found
    for o in distribution.apt_sources:
        if o.site:
            # Loop over mandatory attributes
            for a in ["name", "component", "origin",
                      "label", "site", "archive_uri"]:
                assert getattr(o, a), "A non-local origin needs a " + a
            # Note: date and architecture are not mandatory (and not found on
            # travis)
            break
    else:
        assert False, "A non-local origin must be found"


@pytest.mark.integration
@skip_if_no_apt_cache
def test_check_bin_packages():
    # Gather files in /usr/bin and /usr/lib
    files = list_all_files("/usr/bin") + list_all_files("/usr/lib")
    tracer = DebTracer()
    distributions = list(tracer.identify_distributions(files))
    assert len(distributions) == 1
    distribution, unknown_files = distributions[0]
    non_local_origins = [o for o in distribution.apt_sources if o.site]
    assert len(non_local_origins) > 0, "A non-local origin must be found"
    for o in non_local_origins:
        # Loop over mandatory attributes
        for a in ["name", "component", "origin",
                  "label", "site", "archive_uri"]:
            assert getattr(o, a), "A non-local origin needs a " + a
# Allow bin files to be not associated with a specific package
#    assert len(unknown_files) == 0, "Files not found in packages: " + \
#                                    str(unknown_files)


def list_all_files(dir):
    files = [join(dir, f) for f in os.listdir(dir)
             if (isfile(join(dir, f)) and
                 not islink(join(dir, f)))]
    return files


# def test_find_release_file():
#     fp = lambda p: os.path.join('/var/lib/apt/lists', p)
#
#     def mocked_exists(path):
#         return path in {
#             fp('s_d_d_data_crap_InRelease'),
#             fp('s_d_d_datas_InRelease'),
#             fp('s_d_d_data_InRelease'),
#             fp('s_d_d_sid_InRelease'),
#             fp('s_d_d_InRelease')
#         }
#
#     with mock.patch('os.path.exists', mocked_exists):
#         assert DebTracer._find_release_file(
#             fp('s_d_d_data_non-free_binary-amd64_Packages')) == \
#                fp('s_d_d_data_InRelease')
#         assert DebTracer._find_release_file(
#             fp('s_d_d_data_non-free_binary-i386_Packages')) == \
#                fp('s_d_d_data_InRelease')
#         assert DebTracer._find_release_file(
#             fp('oths_d_d_data_non-free_binary-i386_Packages')) is None


def test_trace_nonexisting_file():
    files = ["/is/not/there/"]
    manager = DebTracer()
    packages, unknown_files = manager.identify_packages_from_files(files)
    # get_details_for_packages doesn't fail on an empty package list.
    assert not packages
    packages = manager.get_details_for_packages(packages)
    assert not packages


def test_utf8_file():
    files = [u"/usr/share/ca-certificates/mozilla/"
             u"TÜBİTAK_UEKAE_Kök_Sertifika_Hizmet_Sağlayıcısı_-_Sürüm_3.crt"]
    manager = DebTracer()
    # Simple sanity check that the pipeline works with utf-8
    (packages, unknown_files) = \
        manager.identify_packages_from_files(files)
    packages = manager.get_details_for_packages(packages)
    # If the file exists, it should be in ca-certificates
    if os.path.isfile(files[0]):
        assert packages[0].name == "ca-certificates"
    else:  # Otherwise just make sure we didn't throw an exception
        assert True


def test_get_packagefields_for_files():
    manager = DebTracer()
    # TODO: mock! and bring back afni and fail2ban
    files = ['/bin/sh',  # the tricky one with alternatives etc, on my system - provided by dash
             '/lib/i386-linux-gnu/libz.so.1.2.8', '/lib/x86_64-linux-gnu/libz.so.1.2.8',  # multiarch
             '/usr/lib/afni/bin/afni',  # from contrib
             '/usr/bin/fail2ban-server', '/usr/bin/fail2ban-server', # arch all and multiple
             '/bogus'
             ]

    def exec_cmd_batch_mock(session, cmd, subfiles, exc_classes):
        assert subfiles == files  # we get all of the passed in
        assert cmd == ['dpkg-query', '-S']

        yield ("""\
diversion by dash from: /bin/sh
diversion by dash to: /bin/sh.distrib
dash: /bin/sh
zlib1g:i386: /lib/i386-linux-gnu/libz.so.1.2.8
zlib1g:amd64: /lib/x86_64-linux-gnu/libz.so.1.2.8
afni: /usr/lib/afni/bin/afni
fail2ban: /usr/bin/fail2ban-server
fail2ban: /usr/bin/fail2ban-server
""", None, None)
    with mock.patch('niceman.distributions.debian.execute_command_batch',
                    exec_cmd_batch_mock):
        out = manager._get_packagefields_for_files(files)

    assert out == {
        '/lib/i386-linux-gnu/libz.so.1.2.8': {'name': u'zlib1g', 'architecture': u'i386'},
        '/lib/x86_64-linux-gnu/libz.so.1.2.8': {'name': u'zlib1g', 'architecture': u'amd64'},
        '/usr/bin/fail2ban-server': {'name': u'fail2ban'},
        '/usr/lib/afni/bin/afni': {'name': u'afni'},
        '/bin/sh': {'name': u'dash'}
    }


def test_parse_dpkgquery_line():
    parse = DebTracer()._parse_dpkgquery_line

    mock_values = {
        "unique": {"name": "pkg",
                   "path": "/path/to/file",
                   "pkgs_rest": None},
        "multi_dir": {"name": "pkg",
                      "path": os.getcwd(),
                      "pkgs_rest": ", more, packages"},
        "multi_file": {"name": "pkg",
                       "path": __file__,
                       "pkgs_rest": ", more, packages"}
    }

    with mock.patch("niceman.distributions.debian.parse_dpkgquery_line",
                    mock_values.get):
        assert parse("unique") == {"name": "pkg",
                                   "path": "/path/to/file"}
        assert parse("multi_dir") is None
        with swallow_logs(new_level=logging.WARNING) as log:
            assert parse("multi_file") == {"name": "pkg",
                                           "path": __file__}
            assert any("multiple packages " in ln for ln in log.lines)


@pytest.fixture
def setup_packages():
    """set up the package comparison tests"""
    a = DEBPackage(name='p1')
    b = DEBPackage(name='p1', version='1.0')
    c = DEBPackage(name='p1', version='1.1')
    d = DEBPackage(name='p1', architecture='i386')
    e = DEBPackage(name='p1', architecture='alpha')
    f = DEBPackage(name='p1', version='1.1', architecture='i386')
    g = DEBPackage(name='p2')
    return (a, b, c, d, e, f, g)

def test_package_satisfies(setup_packages):
    (p1, p1v10, p1v11, p1ai, p1aa, p1v11ai, p2) = setup_packages
    assert p1.satisfies(p1)
    assert p1v10.satisfies(p1v10)
    assert not p1.satisfies(p1v10)
    assert p1v10.satisfies(p1)
    assert not p1v10.satisfies(p1v11)
    assert not p1.satisfies(p2)
    assert not p1v10.satisfies(p2)
    assert not p2.satisfies(p1v10)
    assert not p1v10.satisfies(p1aa)
    assert p1aa.satisfies(p1)
    assert not p1aa.satisfies(p1v10)
    assert not p1aa.satisfies(p1ai)
    assert not p1v11.satisfies(p1v11ai)
    assert p1v11ai.satisfies(p1v11)

@pytest.fixture
def setup_distributions(setup_packages):
    (p1, p1v10, p1v11, p1ai, p1aa, p1v11ai, p2) = setup_packages
    d1 = DebianDistribution(name='debian 1')
    d1.packages = [p1]
    d2 = DebianDistribution(name='debian 2')
    d2.packages = [p1v11]
    return (d1, d2)

def test_distribution_satisfies_package(setup_distributions, setup_packages):
    (d1, d2) = setup_distributions
    (p1, p1v10, p1v11, p1ai, p1aa, p1v11ai, p2) = setup_packages
    assert d1.satisfies_package(p1)
    assert not d1.satisfies_package(p1v10)
    assert d2.satisfies_package(p1)
    assert not d2.satisfies_package(p1v10)
    assert d2.satisfies_package(p1v11)

def test_distribution_statisfies(setup_distributions):
    (d1, d2) = setup_distributions
    assert not d1.satisfies(d2)
    assert d2.satisfies(d1)

def test_distribution_sub(setup_packages):
    (p1, p1v10, p1v11, p1ai, p1aa, p1v11ai, p2) = setup_packages
    d1 = DebianDistribution(name='debian 1')
    d1.packages = [p1, p2]
    d2 = DebianDistribution(name='debian 2')
    d2.packages = [p1v11, p2]
    assert d1-d2 == []
    result = d2-d1
    assert len(result) == 1
    assert result[0] == p1v11



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from niceman.cmdline.main import main
from niceman.formats import Provenance

import logging

from niceman.utils import swallow_logs, swallow_outputs, make_tempfile
from niceman.tests.utils import assert_in, skip_if_no_apt_cache

from ..retrace import identify_distributions

def test_retrace(reprozip_spec2):
    """
    Test installing packages on the localhost.
    """
    with swallow_logs(new_level=logging.DEBUG) as log:
        args = ['retrace',
                '--spec', reprozip_spec2,
                ]
        main(args)
        assert_in("reading spec file " + reprozip_spec2, log.lines)


def test_retrace_to_output_file(reprozip_spec2):
    with make_tempfile() as outfile:
        args = ['retrace',
                '--spec', reprozip_spec2,
                '--output-file', outfile]
        main(args)

        ## Perform a simple check of whether the output file can be
        ## loaded.
        provenance = Provenance.factory(outfile)
        assert len(provenance.get_distributions()) == 1


@skip_if_no_apt_cache
def test_retrace_normalize_paths():
    # Retrace should normalize paths before passing them to tracers.
    with swallow_outputs() as cm:
        main(["retrace", "/sbin/../sbin/iptables"])
        assert "name: debian" in cm.out


def get_tracer_session(protocols):
    class FakeSession(object):
        """A fake session attributes and methods of which should not
        actually be used only but isdir.
        If anything else is accessed, it means that we have some assumptions
        """

        def isdir(self, _):
            return False  # TODO: make it parametric

    tracer_classes = []
    for itracer, protocol in enumerate(protocols):
        # Test the loop logic
        class FakeTracer(object):
            _protocol = protocol[:]
            HANDLES_DIRS = False  # ???

            def __init__(self, session):
                assert session
                assert self._protocol, \
                    "No more protocols to go through, but were were asked to"
                self._current_protocol = self._protocol.pop(0)

            def identify_distributions(self, files):
                for item in self._current_protocol:
                    yield item
        FakeTracer.__name__ = "FakeTracer%d" % itracer
        tracer_classes.append(FakeTracer)
    return tracer_classes, FakeSession()


def _check_loop_protocol(protocols, files, tenvs, tfiles):
    tracer_classes, session = get_tracer_session(protocols)
    dists, unknown_files = identify_distributions(
        files, session, tracer_classes=tracer_classes)
    assert not any(t._protocol for t in tracer_classes), "we exhausted the protocol"
    assert dists == tenvs
    assert unknown_files == tfiles


def test_retrace_loop_over_tracers():
    _check_loop_protocol(
        [  # Tracers
            [  # Tracer passes
                [  # what to yield
                    ("Env1", {"thefile"})
                ]
            ]
        ],
        files=["thefile"],
        tenvs=['Env1'],
        tfiles={"thefile"})

    # The 2nd tracer consumes everything
    _check_loop_protocol(
        [  # Tracers
            [  # Tracer passes
                [  # what to yield
                    ("Env1", {"thefile"})
                ],
            ],
            [  # Tracer passes
                [  # what to yield
                    ("Env2", set())
                ],
            ]
        ],
        files=["thefile"],
        tenvs=['Env1', 'Env2'],
        tfiles=set())

    # The fancy multi-yield and producing stuff
    _check_loop_protocol(
        [  # Tracers
            [  # Tracer passes
                [  # what to yield
                    ("Env1", {"file2", "file3"}),
                ],
                [
                    ("Env3", {"file3"})  # consume file4
                ],
                []  # finale
            ],
            [  # Tracer passes
                [  # what to yield
                    ("Env2", {"file3", "file4", "file5"}),
                    ("Env2.1", {"file3", "file4"})
                ],
                [

                ],
                []  # finale
            ]
        ],
        files=["file1", "file2"],
        tenvs=['Env1', 'Env2', 'Env2.1', 'Env3'],
        tfiles={'file3'})


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Report if a specification satisfies the requirements in another specification
"""

import sys
import time

from .base import Interface
from ..support.constraints import EnsureStr
from ..support.exceptions import InsufficientArgumentsError
from ..support.param import Parameter
from niceman.formats.niceman import NicemanProvenance
from ..distributions.debian import DebianDistribution
from ..distributions.conda import CondaDistribution
from ..distributions.vcs import GitDistribution, SVNDistribution

__docformat__ = 'restructuredtext'

from logging import getLogger
lgr = getLogger('niceman.api.retrace')


class MultipleDistributionsError(Exception):
    """Multiple distributions of a given type found"""

    def __init__(self, cls):
        self.cls = cls


class Diff(Interface):
    """Report if a specification satisfies the requirements in another 
    specification

    Examples
    --------

      $ niceman diff --env environment.yml --req required_environment.yml

    """

    _params_ = dict(
        prov1=Parameter(
            doc="NICEMAN provenance file", 
            metavar='prov1',
            constraints=EnsureStr()),
        prov2=Parameter(
            metavar="prov2",
            doc="NICEMAN provenance file", 
            constraints=EnsureStr())
    )

    @staticmethod
    def __call__(prov1, prov2):

        env_1 = NicemanProvenance(prov1).get_environment()
        env_2 = NicemanProvenance(prov2).get_environment()

        for (dist_type, pkg_type) in ((DebianDistribution, "Debian package"), 
                                      (CondaDistribution, "Conda package")):

            dist_1 = env_1.get_distribution(dist_type)
            if dist_1:
                pkgs_1 = { p._cmp_id: p for p in dist_1.packages }
            else:
                pkgs_1 = {}
            dist_2 = env_2.get_distribution(dist_type)
            if dist_2:
                pkgs_2 = { p._cmp_id: p for p in dist_2.packages }
            else:
                pkgs_2 = {}

            pkgs_1_s = set(pkgs_1)
            pkgs_2_s = set(pkgs_2)
    
            pkgs_only_1 = pkgs_1_s - pkgs_2_s
            pkgs_only_2 = pkgs_2_s - pkgs_1_s
    
            if pkgs_only_1 or pkgs_only_2:
                print(pkg_type + 's:')
            if pkgs_only_1:
                for cmp_key in sorted(pkgs_only_1):
                    package = pkgs_1[cmp_key]
                    print('< %s %s' % (" ".join(cmp_key), package.version))
            if pkgs_only_2 and pkgs_only_2:
                print('---')
            if pkgs_only_2:
                for cmp_key in sorted(pkgs_only_2):
                    package = pkgs_2[cmp_key]
                    print('> %s %s' % (" ".join(cmp_key), package.version))
    
            for cmp_key in pkgs_1_s.intersection(pkgs_2_s):
                package_1 = pkgs_1[cmp_key]
                package_2 = pkgs_2[cmp_key]
                if package_1.version != package_2.version:
                    print('%s %s:' % (pkg_type, " ".join(cmp_key)))
                    print('< %s' % package_1.version)
                    print('---')
                    print('> %s' % package_2.version)

        for (dist_type, repo_type) in ((GitDistribution, 'Git'), 
                                       (SVNDistribution, 'SVN')):

            dist_1 = env_1.get_distribution(dist_type)
            if dist_1:
                repos_1 = { p.identifier: p for p in dist_1.packages }
            else:
                repos_1 = {}
            dist_2 = env_2.get_distribution(dist_type)
            if dist_2:
                repos_2 = { p.identifier: p for p in dist_2.packages }
            else:
                repos_2 = {}

            repos_1_s = set(repos_1)
            repos_2_s = set(repos_2)

            repos_1_only = repos_1_s - repos_2_s
            repos_2_only = repos_2_s - repos_1_s

            if repos_1_only or repos_2_only:
                print('%s repositories:' % repo_type)
            if repos_1_only:
                for repo_id in repos_1_only:
                    repo = repos_1[repo_id]
                    print('< %s (%s)' % (repo.identifier, repo.path))
            if repos_1_only and repos_2_only:
                print('---')
            if repos_2_only:
                for repo_id in repos_2_only:
                    repo = repos_2[repo_id]
                    print('> %s (%s)' % (repo.identifier, repo.path))

            for repo_id in repos_1_s.intersection(repos_2_s):
                repo_1 = repos_1[repo_id]
                repo_2 = repos_2[repo_id]
                if repo_1.commit == repo_2.commit:
                    continue
                print('%s repository %s:' % (repo_type, repo_id))
                print('< %s (%s)' % (repo_1.commit, repo_1.path))
                print('---')
                print('> %s (%s)' % (repo_2.commit, repo_2.path))

        files1 = set(env_1.files)
        files2 = set(env_2.files)
    
        files_1_only = files1 - files2
        files_2_only = files2 - files1
    
        if files_1_only or files_2_only:
            print('Files:')
            for fname in files_1_only:
                print('< %s' % fname)
            if files_1_only and files_2_only:
                print('---')
            for fname in files_2_only:
                print('> %s' % fname)

        return



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""" niceman constants
"""

# whenever to come - to be defined here

# Define a RepDigest for the testing SSH Docker container for use across
# all testing modules. Using a tag name does not guarantee the same image will
# be used over time.
TEST_SSH_DOCKER_DIGEST = 'rastasheep/ubuntu-sshd@sha256:918aae46c217701b5516776e0ccc9ebb93abce5ebf3efa4bfd75a842cffc4e04'
